# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2015-2016, Lars Asplund lars.anders.asplund@gmail.com

"""
A general tokenizer
"""

import collections
import re
from vunit.ostools import read_file, file_exists, simplify_path

TokenType = collections.namedtuple("Token", ["kind", "value", "location"])


def Token(kind, value="", location=None):  # pylint: disable=invalid-name
    return TokenType(kind, value, location)


def new_token_kind(name):
    """
    Create a new token kind with nice __repr__
    """

    def new_token(kind, value='', location=None):
        """
        Create new token of kind
        """
        return Token(kind, value, location)

    cls = type(name,
               (object,), {
                   "__repr__": lambda self: name,
                   "__call__": new_token
               })
    return cls()


class Tokenizer(object):
    """
    Maintain a prioritized list of token regex
    """

    def __init__(self):
        self._regexs = []
        self._assoc = {}
        self._regex = None

    def add(self, kind, regex, func=None):
        """
        Add token type
        """
        key = chr(ord('a') + len(self._regexs))
        self._regexs.append((key, regex))
        self._assoc[key] = (kind, func)
        return kind

    def finalize(self):
        self._regex = re.compile("|".join("(?P<%s>%s)" % spec for spec in self._regexs), re.VERBOSE | re.MULTILINE)

    def tokenize(self, code, file_name=None, previous_location=None, create_locations=False):
        """
        Tokenize the code
        """
        tokens = []
        start = 0
        while True:
            match = self._regex.search(code, pos=start)
            if match is None:
                break
            lexpos = (start, match.end() - 1)
            start = match.end()
            kind, func = self._assoc[match.lastgroup]
            value = match.group(match.lastgroup)

            if create_locations:
                location = ((file_name, lexpos), previous_location)
            else:
                location = None

            token = Token(kind, value, location)
            if func is not None:
                token = func(token)

            if token is not None:
                tokens.append(token)
        return tokens


class TokenStream(object):
    """
    Helper class for traversing a stream of tokens
    """

    def __init__(self, tokens):
        self._tokens = tokens
        self._idx = 0

    def __len__(self):
        return len(self._tokens)

    def __getitem__(self, index):
        return self._tokens[index]

    @property
    def eof(self):
        return not self._idx < len(self._tokens)

    @property
    def idx(self):
        return self._idx

    @property
    def current(self):
        return self._tokens[self._idx]

    def peek(self, offset=0):
        return self._tokens[self._idx + offset]

    def skip_while(self, *kinds):
        """
        Skip forward while token kind is present
        """
        while not self.eof:
            if not any(self._tokens[self._idx].kind == kind for kind in kinds):
                break
            self._idx += 1
        return self._idx

    def skip_until(self, *kinds):
        """
        Skip forward until token kind is present
        """
        while not self.eof:
            if any(self._tokens[self._idx].kind == kind for kind in kinds):
                break
            self._idx += 1
        return self._idx

    def pop(self):
        """
        Return current token and advance stream
        """
        if self.eof:
            raise EOFException()

        self._idx += 1
        return self._tokens[self._idx - 1]

    def expect(self, *kinds):
        """
        Expect to pop token with any of kinds
        """
        token = self.pop()
        if token.kind not in kinds:
            if len(kinds) == 1:
                expected = str(kinds[0])
            else:
                expected = "any of [%s]" % ", ".join(str(kind) for kind in kinds)
            raise LocationException.error("Expected %s got %s" % (expected, token.kind),
                                          token.location)
        return token

    def slice(self, start, end):
        return self._tokens[start:end]


def describe_location(location, first=True):
    """
    Describe the location as a string
    """
    if location is None:
        return "Unknown location"

    ((file_name, (start, end)), previous) = location

    retval = ""
    if previous is not None:
        retval += describe_location(previous, first=False) + "\n"

    if file_name is None:
        retval += "Unknown Python string"
        return retval

    if not file_exists(file_name):
        retval += "Unknown location in %s" % file_name
        return retval

    contents = read_file(file_name)

    if first:
        prefix = "at"
    else:
        prefix = "from"

    count = 0
    for lineno, line in enumerate(contents.splitlines()):
        lstart = count
        lend = lstart + len(line)
        if lstart <= start and start <= lend:
            retval += "%s %s line %i:\n" % (prefix, simplify_path(file_name), lineno + 1)
            retval += line + "\n"
            retval += (" " * (start - lstart)) + ("~" * (min(lend - 1, end) - start + 1))
            return retval

        count = lend + 1
    return retval


class EOFException(Exception):
    """
    End of file
    """
    pass


class LocationException(Exception):
    """
    A an exception to be raised when there is a problem in a location
    """
    @classmethod
    def error(cls, message, location):
        return cls(message, location, "error")

    @classmethod
    def warning(cls, message, location):
        return cls(message, location, "warning")

    @classmethod
    def debug(cls, message, location):
        return cls(message, location, "debug")

    def __init__(self, message, location, severity):
        Exception.__init__(self)
        assert severity in ("debug", "warning", "error")
        self._severtity = severity
        self._message = message
        self._location = location

    def log(self, logger):
        """
        Log the exception
        """
        if self._severtity == "error":
            method = logger.error
        elif self._severtity == "warning":
            method = logger.warning
        else:
            method = logger.debug

        method(self._message + "\n%s",
               describe_location(self._location))


def add_previous(location, previous):
    """
    Add previous location
    """
    if location is None:
        return previous

    current, old_previous = location
    return (current, add_previous(old_previous, previous))


def strip_previous(location):
    """
    Strip previous location
    """
    if location is None:
        return None
    return location[0]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2017, Lars Asplund lars.anders.asplund@gmail.com

"""
Provided functionality to run a suite of test in a robust way
"""


from __future__ import print_function

import os
from os.path import join, exists, abspath, basename, relpath
import traceback
import threading
import sys
import time
import logging
import string
import vunit.ostools as ostools
from vunit.test_report import PASSED, FAILED
from vunit.hashing import hash_string
LOGGER = logging.getLogger(__name__)


class TestRunner(object):  # pylint: disable=too-many-instance-attributes
    """
    Administer the execution of a list of test suites
    """

    VERBOSITY_QUIET = 0
    VERBOSITY_NORMAL = 1
    VERBOSITY_VERBOSE = 2

    def __init__(self, report, output_path,
                 verbosity=VERBOSITY_NORMAL,
                 num_threads=1,
                 dont_catch_exceptions=False):
        self._lock = threading.Lock()
        self._local = threading.local()
        self._report = report
        self._output_path = output_path
        assert verbosity in (self.VERBOSITY_QUIET,
                             self.VERBOSITY_NORMAL,
                             self.VERBOSITY_VERBOSE)
        self._verbosity = verbosity
        self._num_threads = num_threads
        self._stdout = sys.stdout
        self._stderr = sys.stderr
        self._dont_catch_exceptions = dont_catch_exceptions

    @property
    def _is_verbose(self):
        return self._verbosity == self.VERBOSITY_VERBOSE

    @property
    def _is_quiet(self):
        return self._verbosity == self.VERBOSITY_QUIET

    def run(self, test_suites):
        """
        Run a list of test suites
        """

        if not exists(self._output_path):
            os.makedirs(self._output_path)

        self._create_test_mapping_file(test_suites)

        num_tests = 0
        for test_suite in test_suites:
            for test_name in test_suite.test_cases:
                num_tests += 1
                if self._is_verbose:
                    print("Running test: " + test_name)

        if self._is_verbose:
            print("Running %i tests" % num_tests)
            print()

        self._report.set_expected_num_tests(num_tests)

        scheduler = TestScheduler(test_suites)

        threads = []

        # Disable continuous output in parallel mode
        write_stdout = self._is_verbose and self._num_threads == 1

        try:
            sys.stdout = ThreadLocalOutput(self._local, self._stdout)
            sys.stderr = ThreadLocalOutput(self._local, self._stdout)

            # Start P-1 worker threads
            for _ in range(self._num_threads - 1):
                new_thread = threading.Thread(target=self._run_thread,
                                              args=(write_stdout, scheduler, num_tests, False))
                threads.append(new_thread)
                new_thread.start()

            # Run one worker in main thread such that P=1 is not multithreaded
            self._run_thread(write_stdout, scheduler, num_tests, True)

            scheduler.wait_for_finish()

        except KeyboardInterrupt:
            LOGGER.debug("TestRunner: Caught Ctrl-C shutting down")
            ostools.PROGRAM_STATUS.shutdown()
            raise

        finally:
            for thread in threads:
                thread.join()

            sys.stdout = self._stdout
            sys.stderr = self._stderr
            LOGGER.debug("TestRunner: Leaving")

    def _run_thread(self, write_stdout, scheduler, num_tests, is_main):
        """
        Run worker thread
        """
        self._local.output = self._stdout

        while True:
            test_suite = None
            try:
                test_suite = scheduler.next()

                output_path = create_output_path(self._output_path, test_suite.name)
                output_file_name = join(output_path, "output.txt")

                with self._lock:  # pylint: disable=not-context-manager
                    for test_name in test_suite.test_cases:
                        print("Starting %s" % test_name)
                    print("Output file: %s" % relpath(output_file_name))

                self._run_test_suite(test_suite,
                                     write_stdout,
                                     num_tests,
                                     output_path,
                                     output_file_name)

            except StopIteration:
                return

            except KeyboardInterrupt:
                # Only main thread should handle KeyboardInterrupt
                if is_main:
                    raise
                else:
                    return

            finally:
                if test_suite is not None:
                    scheduler.test_done()

    def _run_test_suite(self,
                        test_suite,
                        write_stdout,
                        num_tests,
                        output_path,
                        output_file_name):
        """
        Run the actual test suite
        """
        start_time = ostools.get_time()

        try:
            # If we could not clean output path, fail all tests
            ostools.renew_path(output_path)
            output_file = open(output_file_name, "w")
        except KeyboardInterrupt:
            raise
        except:  # pylint: disable=bare-except
            results = self._fail_suite(test_suite)
            with self._lock:  # pylint: disable=not-context-manager
                traceback.print_exc()
                self._add_results(test_suite, results, start_time, num_tests, output_file_name)
            return

        try:
            if write_stdout:
                self._local.output = TeeToFile([self._stdout, output_file])
            else:
                self._local.output = TeeToFile([output_file])

            results = test_suite.run(output_path)
        except KeyboardInterrupt:
            raise
        except:  # pylint: disable=bare-except
            if self._dont_catch_exceptions:
                raise
            traceback.print_exc()
            results = self._fail_suite(test_suite)
        finally:
            self._local.output = self._stdout
            output_file.flush()
            output_file.close()

        any_not_passed = any(value != PASSED for value in results.values())

        with self._lock:  # pylint: disable=not-context-manager
            if (not write_stdout) and (any_not_passed or self._is_verbose) and not self._is_quiet:
                self._print_output(output_file_name)
            self._add_results(test_suite, results, start_time, num_tests, output_file_name)

    def _create_test_mapping_file(self, test_suites):
        """
        Create a file mapping test name to test output folder.
        This is to allow the user to find the test output folder when it is hashed
        """
        mapping_file_name = join(self._output_path, "test_name_to_path_mapping.txt")

        # Load old mapping to remember non-deleted test folders as well
        # even when re-running only a single test case
        if exists(mapping_file_name):
            with open(mapping_file_name, "r") as fptr:
                mapping = set(fptr.read().splitlines())
        else:
            mapping = set()

        for test_suite in test_suites:
            test_output = create_output_path(self._output_path, test_suite.name)
            mapping.add("%s %s" % (basename(test_output), test_suite.name))

        # Sort by everything except hash
        mapping = sorted(mapping, key=lambda value: value[value.index(" "):])

        with open(mapping_file_name, "w") as fptr:
            for value in mapping:
                fptr.write(value + "\n")

    @staticmethod
    def _print_output(output_file_name):
        """
        Print contents of output file if it exists
        """
        with open(output_file_name, "r") as fread:
            for line in fread:
                print(line, end="")

    def _add_results(self, test_suite, results, start_time, num_tests, output_file_name):
        """
        Add results to test report
        """
        runtime = ostools.get_time() - start_time
        time_per_test = runtime / len(results)

        for test_name in test_suite.test_cases:
            status = results[test_name]
            self._report.add_result(test_name,
                                    status,
                                    time_per_test,
                                    output_file_name)
            self._report.print_latest_status(total_tests=num_tests)
        print()

    @staticmethod
    def _fail_suite(test_suite):
        """ Return failure for all tests in suite """
        results = {}
        for test_name in test_suite.test_cases:
            results[test_name] = FAILED
        return results


class TeeToFile(object):
    """
    Provide a write method which writes to multiple files
    like the unix 'tee' command.
    """
    def __init__(self, files):
        self._files = files

    def write(self, txt):
        for ofile in self._files:
            ofile.write(txt)

    def flush(self):
        for ofile in self._files:
            ofile.flush()


class ThreadLocalOutput(object):
    """
    Replacement for stdout/err that separates re-directs
    output to a thread local file interface
    """
    def __init__(self, local, stdout):
        self._local = local
        self._stdout = stdout

    def write(self, txt):
        """
        Write to file object
        """
        if hasattr(self._local, "output"):
            self._local.output.write(txt)
        else:
            self._stdout.write(txt)

    def flush(self):
        """
        Flush file object
        """
        if hasattr(self._local, "output"):
            self._local.output.flush()
        else:
            self._stdout.flush()


class TestScheduler(object):
    """
    Schedule tests to different treads
    """

    def __init__(self, tests):
        self._lock = threading.Lock()
        self._tests = tests
        self._idx = 0
        self._num_done = 0

    def __iter__(self):
        return self

    def __next__(self):
        """
        Iterator in Python 3
        """
        return self.__next__()

    def next(self):
        """
        Iterator in Python 2
        """
        ostools.PROGRAM_STATUS.check_for_shutdown()
        with self._lock:  # pylint: disable=not-context-manager
            if self._idx < len(self._tests):
                idx = self._idx
                self._idx += 1
                return self._tests[idx]
            else:
                raise StopIteration

    def test_done(self):
        """
        Signal that a test has been done
        """
        with self._lock:  # pylint: disable=not-context-manager
            self._num_done += 1

    def is_finished(self):
        with self._lock:  # pylint: disable=not-context-manager
            return self._num_done >= len(self._tests)

    def wait_for_finish(self):
        """
        Block until all tests have been done
        """
        while not self.is_finished():
            time.sleep(0.05)

LEGAL_CHARS = string.printable
ILLEGAL_CHARS = ' <>"|:*%?\\/#&;()'


def _is_legal(char):
    """
    Return true if the character is legal to have in a file name
    """
    return (char in LEGAL_CHARS) and (char not in ILLEGAL_CHARS)


def create_output_path(output_path, test_suite_name):
    """
    Create the full output path of a test case.
    Ensure no bad characters and no long path names.
    """
    output_path = abspath(output_path)
    safe_name = "".join(char if _is_legal(char) else '_' for char in test_suite_name) + "_"
    hash_name = hash_string(test_suite_name)

    if "VUNIT_SHORT_TEST_OUTPUT_PATHS" in os.environ:
        full_name = hash_name
    elif sys.platform == "win32":
        max_path = 260
        margin = int(os.environ.get("VUNIT_TEST_OUTPUT_PATH_MARGIN", "100"))
        prefix_len = len(output_path)
        full_name = safe_name[:min(max_path - margin - prefix_len - len(hash_name), len(safe_name))] + hash_name
    else:
        full_name = safe_name + hash_name

    return join(output_path, full_name)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2015, Lars Asplund lars.anders.asplund@gmail.com

"""
Templates common to all datatype codecs.
"""

from string import Template


class DatatypeCodecTemplate(object):
    """Base class for datatype codec templates"""

    to_string_declarations = Template("""\
  function to_string (
    constant data : $type)
    return string;

""")


class DatatypeStdCodecTemplate(DatatypeCodecTemplate):
    """Templates when generating standard codecs"""

    codec_declarations = Template("""\
  function encode (
    constant data : $type)
    return string;
  alias encode_$type is encode[$type return string];
  function decode (
    constant code : string)
    return $type;
  alias decode_$type is decode[string return $type];
  procedure decode (
    constant code   : string;
    variable index : inout positive;
    variable result : out $type);
  alias decode_$type is decode[string, positive, $type];

""")


class DatatypeDebugCodecTemplate(DatatypeCodecTemplate):
    """Templates when generating debugging codecs"""

    codec_declarations = Template("""\
  function encode (
    constant data : $type)
    return string;
  alias encode_$type is encode[$type return string];
  function decode (
    constant code : string)
    return $type;
  alias decode_$type is decode[string return $type];

""")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2016, Lars Asplund lars.anders.asplund@gmail.com

"""
A persistent TCL shell to avoid startup overhead in TCL-based simulators
"""

from __future__ import print_function
import threading
import logging
from vunit.ostools import Process

LOGGER = logging.getLogger(__name__)


class PersistentTclShell(object):
    """
    A persistent TCL shell
    """

    def __init__(self, create_process):
        self._processes = {}
        self._lock = threading.Lock()
        self._create_process = create_process

    def _process(self):
        """
        Create the vsim process
        """
        ident = threading.current_thread().ident

        with self._lock:  # pylint: disable=not-context-manager
            try:
                process = self._processes[ident]
                if process.is_alive():
                    return process
            except KeyError:
                pass

            process = self._create_process(ident)
            self._processes[ident] = process

        process.writeline("puts #VUNIT_RETURN")
        try:
            consumer = SilentOutputConsumer()
            process.consume_output(consumer)
        except Process.NonZeroExitCode:
            # Print output if background vsim process startup failed
            LOGGER.error("Failed to start re-usable background process")
            print(consumer.output)
            raise
        return process

    def execute(self, cmd):
        """
        Execute a command to the persistent TCL shell
        """
        process = self._process()
        process.writeline(cmd)
        process.writeline("puts #VUNIT_RETURN")
        process.consume_output(output_consumer)

    def read_var(self, varname):
        """
        Read a variable from the persistent TCL shell
        """
        process = self._process()
        process.writeline("puts #VUNIT_READVAR=${%s}" % varname)
        consumer = ReadVarOutputConsumer()
        process.consume_output(consumer)
        return consumer.var

    def teardown(self):
        """
        Teardown all active processes before shutdown
        """
        with self._lock:  # pylint: disable=not-context-manager
            for proc in self._processes.values():
                if proc.is_alive():
                    proc.writeline("quit -force -code 0")

            for proc in self._processes.values():
                if proc.is_alive():
                    proc.wait()
            self._processes = {}

    def __del__(self):
        self.teardown()


def output_consumer(line):
    """
    Consume output until reaching #VUNIT_RETURN
    """
    if line.endswith("#VUNIT_RETURN"):
        return True

    print(line)
    return None


class SilentOutputConsumer(object):
    """
    Consume output until reaching #VUNIT_RETURN, silent
    """
    def __init__(self):
        self.output = ""

    def __call__(self, line):
        if line.endswith("#VUNIT_RETURN"):
            return True

        self.output += line + "\n"
        return None


class ReadVarOutputConsumer(object):
    """
    Consume output from modelsim and print with indentation
    """
    def __init__(self):
        self.var = None

    def __call__(self, line):
        self.var = line.split("#VUNIT_READVAR=")[-1].strip()
        return True



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2015-2017, Lars Asplund lars.anders.asplund@gmail.com

"""
Interface for GHDL simulator
"""

from __future__ import print_function
import logging
from os.path import exists, join
import os
import subprocess
import shlex
from sys import stdout  # To avoid output catched in non-verbose mode
from vunit.ostools import Process
from vunit.simulator_interface import (SimulatorInterface,
                                       ListOfStringOption)
from vunit.exceptions import CompileError
LOGGER = logging.getLogger(__name__)


class GHDLInterface(SimulatorInterface):
    """
    Interface for GHDL simulator
    """

    name = "ghdl"
    supports_gui_flag = True

    compile_options = [
        ListOfStringOption("ghdl.flags"),
    ]

    sim_options = [
        ListOfStringOption("ghdl.sim_flags"),
        ListOfStringOption("ghdl.elab_flags"),
    ]

    @staticmethod
    def add_arguments(parser):
        """
        Add command line arguments
        """
        group = parser.add_argument_group("ghdl",
                                          description="GHDL specific flags")
        group.add_argument("--gtkwave-fmt", choices=["vcd", "ghw"],
                           default=None,
                           help="Save .vcd or .ghw to open in gtkwave")
        group.add_argument("--gtkwave-args",
                           default="",
                           help="Arguments to pass to gtkwave")

    @classmethod
    def from_args(cls,
                  output_path,  # pylint: disable=unused-argument
                  args):
        """
        Create instance from args namespace
        """
        prefix = cls.find_prefix()
        return cls(prefix=prefix,
                   gui=args.gui,
                   gtkwave_fmt=args.gtkwave_fmt,
                   gtkwave_args=args.gtkwave_args,
                   backend=cls.determine_backend(prefix))

    @classmethod
    def find_prefix_from_path(cls):
        """
        Find first valid ghdl toolchain prefix
        """
        return cls.find_toolchain(["ghdl"])

    def __init__(self, prefix, gui=False, gtkwave_fmt=None, gtkwave_args="", backend="llvm"):
        SimulatorInterface.__init__(self)
        self._prefix = prefix
        self._project = None

        if gui and (not self.find_executable('gtkwave')):
            raise RuntimeError(
                "Cannot find the gtkwave executable in the PATH environment variable. GUI not possible")

        self._gui = gui
        self._gtkwave_fmt = "ghw" if gui and gtkwave_fmt is None else gtkwave_fmt
        self._gtkwave_args = gtkwave_args
        self._backend = backend
        self._vhdl_standard = None

    @staticmethod
    def determine_backend(prefix):
        """
        Determine the GHDL backend
        """
        mapping = {
            "mcode code generator": "mcode",
            "llvm code generator": "llvm",
            "GCC back-end code generator": "gcc"
        }
        output = subprocess.check_output([join(prefix, "ghdl"), "--version"]).decode()
        for name, backend in mapping.items():
            if name in output:
                LOGGER.debug("Detected GHDL %s", name)
                return backend

        LOGGER.error("Could not detect known LLVM backend by parsing 'ghdl --version'")
        print("Expected to find one of %r" % mapping.keys())
        print("== Output of 'ghdl --version'" + ("=" * 60))
        print(output)
        print("=============================" + ("=" * 60))
        raise AssertionError("No known GHDL back-end could be detected from running 'ghdl --version'")

    def _has_output_flag(self):
        """
        Returns if backend supports output flag
        """
        return self._backend in ("llvm", "gcc")

    def setup_library_mapping(self, project):
        """
        Setup library mapping
        """
        self._project = project
        for library in project.get_libraries():
            if not exists(library.directory):
                os.makedirs(library.directory)

        vhdl_standards = set(source_file.get_vhdl_standard()
                             for source_file in project.get_source_files_in_order()
                             if source_file.is_vhdl)

        if not vhdl_standards:
            self._vhdl_standard = '2008'
        elif len(vhdl_standards) != 1:
            raise RuntimeError("GHDL cannot handle mixed VHDL standards, found %r" % list(vhdl_standards))
        else:
            self._vhdl_standard = list(vhdl_standards)[0]

    def compile_source_file_command(self, source_file):
        """
        Returns the command to compile a single source_file
        """
        if source_file.is_vhdl:
            return self.compile_vhdl_file_command(source_file)

        LOGGER.error("Unknown file type: %s", source_file.file_type)
        raise CompileError

    @staticmethod
    def _std_str(vhdl_standard):
        """
        Convert standard to format of GHDL command line flag
        """
        if vhdl_standard == "2002":
            return "02"
        elif vhdl_standard == "2008":
            return "08"
        elif vhdl_standard == "93":
            return "93"
        raise ValueError("Invalid VHDL standard %s" % vhdl_standard)

    def compile_vhdl_file_command(self, source_file):
        """
        Returns the command to compile a vhdl file
        """
        cmd = [join(self._prefix, 'ghdl'), '-a', '--workdir=%s' % source_file.library.directory,
               '--work=%s' % source_file.library.name,
               '--std=%s' % self._std_str(source_file.get_vhdl_standard())]
        for library in self._project.get_libraries():
            cmd += ["-P%s" % library.directory]
        cmd += source_file.compile_options.get("ghdl.flags", [])
        cmd += [source_file.name]
        return cmd

    def _get_sim_command(self, config, output_path):
        """
        Return GHDL simulation command
        """
        cmd = [join(self._prefix, 'ghdl')]
        cmd += ['--elab-run']
        cmd += ['--std=%s' % self._std_str(self._vhdl_standard)]
        cmd += ['--work=%s' % config.library_name]
        cmd += ['--workdir=%s' % self._project.get_library(config.library_name).directory]
        cmd += ['-P%s' % lib.directory for lib in self._project.get_libraries()]

        if self._has_output_flag():
            cmd += ['-o', join(output_path, "%s-%s" % (config.entity_name,
                                                       config.architecture_name))]
        cmd += config.sim_options.get("ghdl.elab_flags", [])
        cmd += [config.entity_name, config.architecture_name]
        cmd += config.sim_options.get("ghdl.sim_flags", [])

        for name, value in config.generics.items():
            cmd += ['-g%s=%s' % (name, value)]

        cmd += ['--assert-level=%s' % config.vhdl_assert_stop_level]

        if config.sim_options.get("disable_ieee_warnings", False):
            cmd += ["--ieee-asserts=disable"]
        return cmd

    def simulate(self,  # pylint: disable=too-many-locals
                 output_path,
                 test_suite_name,
                 config, elaborate_only):
        """
        Simulate with entity as top level using generics
        """
        if not exists(output_path):
            os.makedirs(output_path)

        cmd = self._get_sim_command(config, output_path)

        if elaborate_only:
            cmd += ["--no-run"]

        if self._gtkwave_fmt is not None:
            data_file_name = join(output_path, "wave.%s" % self._gtkwave_fmt)

            if exists(data_file_name):
                os.remove(data_file_name)

            if self._gtkwave_fmt == "ghw":
                cmd += ['--wave=%s' % data_file_name]
            elif self._gtkwave_fmt == "vcd":
                cmd += ['--vcd=%s' % data_file_name]

        else:
            data_file_name = None

        status = True
        try:
            proc = Process(cmd)
            proc.consume_output()
        except Process.NonZeroExitCode:
            status = False

        if self._gui and not elaborate_only:
            cmd = ["gtkwave"] + shlex.split(self._gtkwave_args) + [data_file_name]
            stdout.write("%s\n" % " ".join(cmd))
            subprocess.call(cmd)

        return status



# Topological sorting of a directed ascylic graph
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2016, Lars Asplund lars.anders.asplund@gmail.com

"""
Functionality to compute a dependency graph
"""


class DependencyGraph(object):
    """
    A dependency graph
    """
    def __init__(self):
        self._forward = {}
        self._backward = {}
        self._nodes = []

    def toposort(self):
        """
        Perform a topological sort returning a list of nodes such that
        every node is located after its dependency nodes
        """
        sorted_nodes = []
        self._visit(sorted(self._nodes),
                    dict((key, sorted(values)) for key, values in self._forward.items()),
                    sorted_nodes.append)
        sorted_nodes = list(reversed(sorted_nodes))
        return sorted_nodes

    def add_node(self, node):
        self._nodes.append(node)

    def add_dependency(self, start, end):
        """
        Add a dependency edge between the start and end node such that
        end node depends on the start node
        """
        new_dependency = (start not in self._forward or
                          end not in self._forward[start])

        if start not in self._forward:
            self._forward[start] = set()

        if end not in self._backward:
            self._backward[end] = set()

        self._forward[start].add(end)
        self._backward[end].add(start)

        return new_dependency

    @staticmethod
    def _visit(nodes, graph, callback):
        """
        Follow graph edges starting from the nodes iteratively
        returning all the nodes visited
        """
        def visit(node):
            """
            Visit a single node and all following nodes in the graph
            that have not already been visisted.
            Detects circular dependencies
            """
            if node in path:
                start = path_ordered.index(node)
                raise CircularDependencyException(path_ordered[start:] + [node, ])

            path.add(node)
            path_ordered.append(node)
            if node in graph:
                for other_node in graph[node]:
                    if other_node not in visited:
                        visit(other_node)
            path.remove(node)
            path_ordered.pop()
            visited.add(node)
            callback(node)

        visited = set()
        for node in nodes:
            if node not in visited:
                path = set()
                path_ordered = []
                visit(node)

    def get_dependent(self, nodes):
        """
        Get all nodes which are directly or indirectly dependent on
        the input nodes
        """
        result = set()
        self._visit(nodes, self._forward, result.add)
        return result

    def get_dependencies(self, nodes):
        """
        Get all nodes which are directly or indirectly dependencies of
        the input nodes
        """
        result = set()
        self._visit(nodes, self._backward, result.add)
        return result

    def get_direct_dependencies(self, node):
        """
        Get the direct dependencies of node
        """
        return self._backward.get(node, set())


class CircularDependencyException(Exception):
    """
    Raised when there are circular dependencies
    """

    def __init__(self, path):
        Exception.__init__(self)
        self.path = path

    def __repr__(self):
        return "CircularDependencyException(%r)" % self.path



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2017, Lars Asplund lars.anders.asplund@gmail.com

"""
Test the check preprocessor
"""


import unittest
from vunit.check_preprocessor import CheckPreprocessor


class TestCheckPreprocessor(unittest.TestCase):
    """
    Test the check preprocessor
    """

    def setUp(self):
        self._check_preprocessor = CheckPreprocessor()

    def _verify_result(self, code, expected_result):
        """
        Assert that the code after preprocessing is equal to the expected_result
        """
        result = self._check_preprocessor.run(code, 'foo.vhd')
        self.assertEqual(result, expected_result)

    def test_that_only_relation_checks_are_preprocessed_so_that_parsing_is_simplified(self):
        code = """
check_relation(a /= b);
check(a /= b or c); -- Same as (a /= b) or c. Not a relation
"""
        expected_result = """
check_relation(a /= b, context_msg => %s);
check(a /= b or c); -- Same as (a /= b) or c. Not a relation
""" % make_context_msg('a', '/=', 'b')

        self._verify_result(code, expected_result)

    def test_that_generated_message_is_combined_with_a_custom_message(self):
        code = """
check_relation(age("John") >= 0, "Age must not be negative.");
"""
        expected_result = """
check_relation(age("John") >= 0, "Age must not be negative.", context_msg => %s);
""" % make_context_msg('age("John")', '>=', '0')

        self._verify_result(code, expected_result)

    def test_that_the_top_level_relation_is_correctly_identified_as_the_main_relation(self):
        code = """
check_relation(foo(a > b) = c);
check_relation((a > b) = c);
check_relation(a > (b = c));
check_relation(( (a > b))  );"""

        expected_result = """
check_relation(foo(a > b) = c, context_msg => %s);
check_relation((a > b) = c, context_msg => %s);
check_relation(a > (b = c), context_msg => %s);
check_relation(( (a > b))  , context_msg => %s);""" % (make_context_msg('foo(a > b)', '=', 'c'),
                                                       make_context_msg('(a > b)', '=', 'c'),
                                                       make_context_msg('a', '>', '(b = c)'),
                                                       make_context_msg('a', '>', 'b'))

        self._verify_result(code, expected_result)

    def test_that_parsing_is_not_fooled_by_strings_containing_characters_relevant_to_parsing(self):
        code = """
check_relation(41 = ascii(')'), "Incorrect ascii for ')'");
check_relation(9 = len("Smile :-)"), "Incorrect length of :-) message");
check_relation(8 = len("Heart => <3"), "Incorrect length of <3 message");
check_relation(a = integer'(9), "This wasn't expected");
check_relation(a = std_logic'('1'), "This wasn't expected");
check_relation(a = func('(','x'), "This wasn't expected");
check_relation(a = std_logic'('1'+'1'), "This wasn't expected");
check_relation(39 = ascii('''), "Incorrect ascii for '''");
check_relation(byte'high = 7);"""

        expected_result = """
check_relation(41 = ascii(')'), "Incorrect ascii for ')'", context_msg => %s);
check_relation(9 = len("Smile :-)"), "Incorrect length of :-) message", context_msg => %s);
check_relation(8 = len("Heart => <3"), "Incorrect length of <3 message", context_msg => %s);
check_relation(a = integer'(9), "This wasn't expected", context_msg => %s);
check_relation(a = std_logic'('1'), "This wasn't expected", context_msg => %s);
check_relation(a = func('(','x'), "This wasn't expected", context_msg => %s);
check_relation(a = std_logic'('1'+'1'), "This wasn't expected", context_msg => %s);
check_relation(39 = ascii('''), "Incorrect ascii for '''", context_msg => %s);
check_relation(byte'high = 7, context_msg => %s);""" % (make_context_msg('41', '=', "ascii(')')"),
                                                        make_context_msg('9', '=', 'len("Smile :-)")'),
                                                        make_context_msg('8', '=', 'len("Heart => <3")'),
                                                        make_context_msg('a', '=', "integer'(9)"),
                                                        make_context_msg('a', '=', "std_logic'('1')"),
                                                        make_context_msg('a', '=', "func('(','x')"),
                                                        make_context_msg('a', '=', "std_logic'('1'+'1')"),
                                                        make_context_msg('39', '=', "ascii(''')"),
                                                        make_context_msg("byte'high", '=', '7'))

        self._verify_result(code, expected_result)

    def test_that_parsing_is_not_fooled_by_embedded_comments(self):
        code = """
check_relation(42 = -- Meaning of life :-)
               hex(66));
check_relation(42 = /* Meaning of
                       life :-) */
               hex(66));
"""
        expected_result = """
check_relation(42 = -- Meaning of life :-)
               hex(66), context_msg => %s);
check_relation(42 = /* Meaning of
                       life :-) */
               hex(66), context_msg => %s);
""" % (make_context_msg('42', '=', 'hex(66)'), make_context_msg('42', '=', 'hex(66)'))

        self._verify_result(code, expected_result)

    def test_that_a_call_without_a_relational_operator_raises_an_error(self):
        code = """
check_relation(a + b - c);
"""
        self.assertRaises(SyntaxError, self._check_preprocessor.run, code, 'foo.vhd')

    def test_that_multiple_parameters_can_be_given(self):
        code = """
check_relation(a = b, level => warning);
check_relation(my_checker, a = b);
check_relation(msg => "Error!", expr => a('<') = b);"""

        expected_result = """
check_relation(a = b, level => warning, context_msg => %s);
check_relation(my_checker, a = b, context_msg => %s);
check_relation(msg => "Error!", expr => a('<') = b, context_msg => %s);""" % (make_context_msg('a', '=', 'b'),
                                                                              make_context_msg('a', '=', 'b'),
                                                                              make_context_msg("a('<')", '=', 'b'))

        self._verify_result(code, expected_result)

    def test_that_all_vhdl_2008_relational_operators_are_recognized(self):
        code = """
check_relation(a = b);
check_relation(a /= b);
check_relation(a < b);
check_relation(a <= b);
check_relation(a > b);
check_relation(a >= b);
check_relation(a ?= b);
check_relation(a ?/= b);
check_relation(a ?< b);
check_relation(a ?<= b);
check_relation(a ?> b);
check_relation(a ?>= b);"""

        expected_result = """
check_relation(a = b, context_msg => %s);
check_relation(a /= b, context_msg => %s);
check_relation(a < b, context_msg => %s);
check_relation(a <= b, context_msg => %s);
check_relation(a > b, context_msg => %s);
check_relation(a >= b, context_msg => %s);
check_relation(a ?= b, context_msg => %s);
check_relation(a ?/= b, context_msg => %s);
check_relation(a ?< b, context_msg => %s);
check_relation(a ?<= b, context_msg => %s);
check_relation(a ?> b, context_msg => %s);
check_relation(a ?>= b, context_msg => %s);""" % (make_context_msg("a", '=', 'b'),
                                                  make_context_msg("a", '/=', 'b'),
                                                  make_context_msg("a", '<', 'b'),
                                                  make_context_msg("a", '<=', 'b'),
                                                  make_context_msg("a", '>', 'b'),
                                                  make_context_msg("a", '>=', 'b'),
                                                  make_context_msg("a", '?=', 'b'),
                                                  make_context_msg("a", '?/=', 'b'),
                                                  make_context_msg("a", '?<', 'b'),
                                                  make_context_msg("a", '?<=', 'b'),
                                                  make_context_msg("a", '?>', 'b'),
                                                  make_context_msg("a", '?>=', 'b'))
        self._verify_result(code, expected_result)


def make_context_msg(left, relation, right):
    return ('"Expected %s %s %s. Left is " & to_string(%s) & ". Right is " & to_string(%s) & "."' %
            (left.replace('"', '""'),
             relation,
             right.replace('"', '""'),
             left,
             right))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2015, Lars Asplund lars.anders.asplund@gmail.com

"""
Module containing the CodecVHDLEnumerationType class.
"""
from string import Template
from vunit.vhdl_parser import VHDLEnumerationType
from vunit.com.codec_datatype_template import DatatypeStdCodecTemplate, DatatypeDebugCodecTemplate


class CodecVHDLEnumerationType(VHDLEnumerationType):
    """Class derived from VHDLEnumerationType to provide codec generator functionality for the enumerated type."""
    def generate_codecs_and_support_functions(self, offset=0, debug=False):
        """Generate codecs and communication support functions for the enumerated type."""
        if not debug:
            template = EnumerationStdCodecTemplate()
        else:
            template = EnumerationDebugCodecTemplate()

        declarations = ''
        definitions = ''

        if len(self.literals) > 256:
            raise NotImplementedError('Support for enums with more than 256 values are yet to be implemented')

        declarations += template.codec_declarations.substitute(type=self.identifier)
        definitions += template.enumeration_codec_definitions.substitute(type=self.identifier, offset=offset)
        declarations += template.to_string_declarations.substitute(type=self.identifier)
        definitions += template.enumeration_to_string_definitions.substitute(type=self.identifier)

        return declarations, definitions


class EnumerationCodecTemplate(object):
    """This class contains enumeration templates common to both standard and debug codecs."""

    enumeration_to_string_definitions = Template("""\
  function to_string (
    constant data : $type)
    return string is
  begin
    return $type'image(data);
  end function to_string;

""")


class EnumerationStdCodecTemplate(DatatypeStdCodecTemplate, EnumerationCodecTemplate):
    """This class contains standard enumeration codec templates."""

    enumeration_codec_definitions = Template("""\
  function encode (
    constant data : $type)
    return string is
    constant offset : natural := $offset;
  begin
    return (1 => character'val($type'pos(data) + offset));
  end function encode;

  procedure decode (
    constant code   : string;
    variable index : inout   positive;
    variable result : out $type) is
    constant offset : natural := $offset;
  begin
    result := $type'val(character'pos(code(index)) - offset);
    index := index + 1;
  end procedure decode;

  function decode (
    constant code : string)
    return $type is
    variable ret_val : $type;
    variable index : positive := code'left;
  begin
    decode(code, index, ret_val);

    return ret_val;
  end function decode;

""")


class EnumerationDebugCodecTemplate(DatatypeDebugCodecTemplate, EnumerationCodecTemplate):
    """This class contains debug enumeration codec templates."""

    enumeration_codec_definitions = Template("""\
  function encode (
    constant data : $type)
    return string is
  begin
    return to_string(data);
  end function encode;

  function decode (
    constant code : string)
    return $type is
  begin
    return $type'value(code);
  end function decode;

""")

    enumeration_to_string_definitions = Template("""\
  function to_string (
    constant data : $type)
    return string is
  begin
    return $type'image(data);
  end function to_string;

""")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2016, Lars Asplund lars.anders.asplund@gmail.com

"""
Verilog tokens
"""

from vunit.parsing.tokenizer import new_token_kind


def __token(kind):
    """
    Create non keyword token kind
    """
    return new_token_kind(kind)


def __keyword(kind):
    """
    Create keyword token kind
    """
    token_kind = new_token_kind(kind)
    KEYWORDS[kind] = token_kind
    return token_kind

KEYWORDS = {}

PREPROCESSOR = __token("preprocessor")
STRING = __token("string")
COMMENT = __token("comment")
IDENTIFIER = __token("identifier")
ESCAPED_NEWLINE = __token("escaped_newline")
NEWLINE = __token("newline")
WHITESPACE = __token("whitespace")
MULTI_COMMENT = __token("multi_comment")
COLON = __token("colon")
SEMI_COLON = __token("semi_colon")
DOUBLE_COLON = __token("double_colon")
HASH = __token("hash")
EQUAL = __token("equal")
LPAR = __token("lpar")
RPAR = __token("rpar")
LBRACKET = __token("lbracket")
RBRACKET = __token("rbracket")
LBRACE = __token("lbrace")
RBRACE = __token("rbrace")
COMMA = __token("comma")
OTHER = __token("other")

MODULE = __keyword("module")
ENDMODULE = __keyword("endmodule")
PACKAGE = __keyword("package")
IMPORT = __keyword("import")
ENDPACKAGE = __keyword("endpackage")
PARAMETER = __keyword("parameter")
ACCEPT_ON = __keyword("accept_on")
ALIAS = __keyword("alias")
ALWAYS = __keyword("always")
ALWAYS_COMB = __keyword("always_comb")
ALWAYS_FF = __keyword("always_ff")
ALWAYS_LATCH = __keyword("always_latch")
AND = __keyword("and")
ASSERT = __keyword("assert")
ASSIGN = __keyword("assign")
ASSUME = __keyword("assume")
AUTOMATIC = __keyword("automatic")
BEFORE = __keyword("before")
BEGIN = __keyword("begin")
BIND = __keyword("bind")
BINS = __keyword("bins")
BINSOF = __keyword("binsof")
BIT = __keyword("bit")
BREAK = __keyword("break")
BUF = __keyword("buf")
BUFIF0 = __keyword("bufif0")
BUFIF1 = __keyword("bufif1")
BYTE = __keyword("byte")
CASE = __keyword("case")
CASEX = __keyword("casex")
CASEZ = __keyword("casez")
CELL = __keyword("cell")
CHANDLE = __keyword("chandle")
CHECKER = __keyword("checker")
CLASS = __keyword("class")
CLOCKING = __keyword("clocking")
CMOS = __keyword("cmos")
CONFIG = __keyword("config")
CONST = __keyword("const")
CONSTRAINT = __keyword("constraint")
CONTEXT = __keyword("context")
CONTINUE = __keyword("continue")
COVER = __keyword("cover")
COVERGROUP = __keyword("covergroup")
COVERPOINT = __keyword("coverpoint")
CROSS = __keyword("cross")
DEASSIGN = __keyword("deassign")
DEFAULT = __keyword("default")
DEFPARAM = __keyword("defparam")
DESIGN = __keyword("design")
DISABLE = __keyword("disable")
DIST = __keyword("dist")
DO = __keyword("do")
EDGE = __keyword("edge")
ELSE = __keyword("else")
END = __keyword("end")
ENDCASE = __keyword("endcase")
ENDCHECKER = __keyword("endchecker")
ENDCLASS = __keyword("endclass")
ENDCLOCKING = __keyword("endclocking")
ENDCONFIG = __keyword("endconfig")
ENDFUNCTION = __keyword("endfunction")
ENDGENERATE = __keyword("endgenerate")
ENDGROUP = __keyword("endgroup")
ENDINTERFACE = __keyword("endinterface")
ENDMODULE = __keyword("endmodule")
ENDPACKAGE = __keyword("endpackage")
ENDPRIMITIVE = __keyword("endprimitive")
ENDPROGRAM = __keyword("endprogram")
ENDPROPERTY = __keyword("endproperty")
ENDSPECIFY = __keyword("endspecify")
ENDSEQUENCE = __keyword("endsequence")
ENDTABLE = __keyword("endtable")
ENDTASK = __keyword("endtask")
ENUM = __keyword("enum")
EVENT = __keyword("event")
EVENTUALLY = __keyword("eventually")
EXPECT = __keyword("expect")
EXPORT = __keyword("export")
EXTENDS = __keyword("extends")
EXTERN = __keyword("extern")
FINAL = __keyword("final")
FIRST_MATCH = __keyword("first_match")
FOR = __keyword("for")
FORCE = __keyword("force")
FOREACH = __keyword("foreach")
FOREVER = __keyword("forever")
FORK = __keyword("fork")
FORKJOIN = __keyword("forkjoin")
FUNCTION = __keyword("function")
GENERATE = __keyword("generate")
GENVAR = __keyword("genvar")
GLOBAL = __keyword("global")
HIGHZ0 = __keyword("highz0")
HIGHZ1 = __keyword("highz1")
IF = __keyword("if")
IFF = __keyword("iff")
IFNONE = __keyword("ifnone")
IGNORE_BINS = __keyword("ignore_bins")
ILLEGAL_BINS = __keyword("illegal_bins")
IMPLEMENTS = __keyword("implements")
IMPLIES = __keyword("implies")
IMPORT = __keyword("import")
INCDIR = __keyword("incdir")
INCLUDE = __keyword("include")
INITIAL = __keyword("initial")
INOUT = __keyword("inout")
INPUT = __keyword("input")
INSIDE = __keyword("inside")
INSTANCE = __keyword("instance")
INT = __keyword("int")
INTEGER = __keyword("integer")
INTERCONNECT = __keyword("interconnect")
INTERFACE = __keyword("interface")
INTERSECT = __keyword("intersect")
JOIN = __keyword("join")
JOIN_ANY = __keyword("join_any")
JOIN_NONE = __keyword("join_none")
LARGE = __keyword("large")
LET = __keyword("let")
LIBLIST = __keyword("liblist")
LIBRARY = __keyword("library")
LOCAL = __keyword("local")
LOCALPARAM = __keyword("localparam")
LOGIC = __keyword("logic")
LONGINT = __keyword("longint")
MACROMODULE = __keyword("macromodule")
MATCHES = __keyword("matches")
MEDIUM = __keyword("medium")
MODPORT = __keyword("modport")
MODULE = __keyword("module")
NAND = __keyword("nand")
NEGEDGE = __keyword("negedge")
NETTYPE = __keyword("nettype")
NEW = __keyword("new")
NEXTTIME = __keyword("nexttime")
NMOS = __keyword("nmos")
NOR = __keyword("nor")
NOSHOWCANCELLED = __keyword("noshowcancelled")
NOT = __keyword("not")
NOTIF0 = __keyword("notif0")
NOTIF1 = __keyword("notif1")
NULL = __keyword("null")
OR = __keyword("or")
OUTPUT = __keyword("output")
PACKAGE = __keyword("package")
PACKED = __keyword("packed")
PARAMETER = __keyword("parameter")
PMOS = __keyword("pmos")
POSEDGE = __keyword("posedge")
PRIMITIVE = __keyword("primitive")
PRIORITY = __keyword("priority")
PROGRAM = __keyword("program")
PROPERTY = __keyword("property")
PROTECTED = __keyword("protected")
PULL0 = __keyword("pull0")
PULL1 = __keyword("pull1")
PULLDOWN = __keyword("pulldown")
PULLUP = __keyword("pullup")
PULSESTYLE_ONDETECT = __keyword("pulsestyle_ondetect")
PULSESTYLE_ONEVENT = __keyword("pulsestyle_onevent")
PURE = __keyword("pure")
RAND = __keyword("rand")
RANDC = __keyword("randc")
RANDCASE = __keyword("randcase")
RANDSEQUENCE = __keyword("randsequence")
RCMOS = __keyword("rcmos")
REAL = __keyword("real")
REALTIME = __keyword("realtime")
REF = __keyword("ref")
REG = __keyword("reg")
REJECT_ON = __keyword("reject_on")
RELEASE = __keyword("release")
REPEAT = __keyword("repeat")
RESTRICT = __keyword("restrict")
RETURN = __keyword("return")
RNMOS = __keyword("rnmos")
RPMOS = __keyword("rpmos")
RTRAN = __keyword("rtran")
RTRANIF0 = __keyword("rtranif0")
RTRANIF1 = __keyword("rtranif1")
S_ALWAYS = __keyword("s_always")
S_EVENTUALLY = __keyword("s_eventually")
S_NEXTTIME = __keyword("s_nexttime")
S_UNTIL = __keyword("s_until")
S_UNTIL_WITH = __keyword("s_until_with")
SCALARED = __keyword("scalared")
SEQUENCE = __keyword("sequence")
SHORTINT = __keyword("shortint")
SHORTREAL = __keyword("shortreal")
SHOWCANCELLED = __keyword("showcancelled")
SIGNED = __keyword("signed")
SMALL = __keyword("small")
SOFT = __keyword("soft")
SOLVE = __keyword("solve")
SPECIFY = __keyword("specify")
SPECPARAM = __keyword("specparam")
STATIC = __keyword("static")
STRING = __keyword("string")
STRONG = __keyword("strong")
STRONG0 = __keyword("strong0")
STRONG1 = __keyword("strong1")
STRUCT = __keyword("struct")
SUPER = __keyword("super")
SUPPLY0 = __keyword("supply0")
SUPPLY1 = __keyword("supply1")
SYNC_ACCEPT_ON = __keyword("sync_accept_on")
SYNC_REJECT_ON = __keyword("sync_reject_on")
TABLE = __keyword("table")
TAGGED = __keyword("tagged")
TASK = __keyword("task")
THIS = __keyword("this")
THROUGHOUT = __keyword("throughout")
TIME = __keyword("time")
TIMEPRECISION = __keyword("timeprecision")
TIMEUNIT = __keyword("timeunit")
TRAN = __keyword("tran")
TRANIF0 = __keyword("tranif0")
TRANIF1 = __keyword("tranif1")
TRI = __keyword("tri")
TRI0 = __keyword("tri0")
TRI1 = __keyword("tri1")
TRIAND = __keyword("triand")
TRIOR = __keyword("trior")
TRIREG = __keyword("trireg")
TYPE = __keyword("type")
TYPEDEF = __keyword("typedef")
UNION = __keyword("union")
UNIQUE = __keyword("unique")
UNIQUE0 = __keyword("unique0")
UNSIGNED = __keyword("unsigned")
UNTIL = __keyword("until")
UNTIL_WITH = __keyword("until_with")
UNTYPED = __keyword("untyped")
USE = __keyword("use")
UWIRE = __keyword("uwire")
VAR = __keyword("var")
VECTORED = __keyword("vectored")
VIRTUAL = __keyword("virtual")
VOID = __keyword("void")
WAIT = __keyword("wait")
WAIT_ORDER = __keyword("wait_order")
WAND = __keyword("wand")
WEAK = __keyword("weak")
WEAK0 = __keyword("weak0")
WEAK1 = __keyword("weak1")
WHILE = __keyword("while")
WILDCARD = __keyword("wildcard")
WIRE = __keyword("wire")
WITH = __keyword("with")
WITHIN = __keyword("within")
WOR = __keyword("wor")
XNOR = __keyword("xnor")
XOR = __keyword("xor")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2016, Lars Asplund lars.anders.asplund@gmail.com

"""
VHDL parsing functionality
"""

import re
from os.path import abspath
import logging
from vunit.cached import cached
from vunit.parsing.encodings import HDL_FILE_ENCODING
LOGGER = logging.getLogger(__name__)


class VHDLParser(object):
    """
    Parse a single VHDL file, caching the result to a database if available
    """

    def __init__(self, database=None):
        self._database = database

    def parse(self, file_name):
        """
        Parse the VHDL code and return a VHDLDesignFile parse result
        parse result is re-used if content hash found in database
        """
        file_name = abspath(file_name)
        return cached("CachedVHDLParser.parse",
                      VHDLDesignFile.parse,
                      file_name,
                      encoding=HDL_FILE_ENCODING,
                      database=self._database)


class VHDLDesignFile(object):  # pylint: disable=too-many-instance-attributes
    """
    Contains VHDL objects found within a file
    """
    def __init__(self,  # pylint: disable=too-many-arguments
                 entities=None,
                 packages=None,
                 package_bodies=None,
                 architectures=None,
                 contexts=None,
                 component_instantiations=None,
                 configurations=None,
                 references=None):
        self.entities = [] if entities is None else entities
        self.packages = [] if packages is None else packages
        self.package_bodies = [] if package_bodies is None else package_bodies
        self.architectures = [] if architectures is None else architectures
        self.contexts = [] if contexts is None else contexts
        self.component_instantiations = [] if component_instantiations is None else component_instantiations
        self.configurations = [] if configurations is None else configurations
        self.references = [] if references is None else references

    @classmethod
    def parse(cls, code):
        """
        Return a new VHDLDesignFile instance by parsing the code
        """
        code = remove_comments(code).lower()
        return cls(entities=list(VHDLEntity.find(code)),
                   architectures=list(VHDLArchitecture.find(code)),
                   packages=list(VHDLPackage.find(code)),
                   package_bodies=list(VHDLPackageBody.find(code)),
                   contexts=list(VHDLContext.find(code)),
                   component_instantiations=list(cls._find_component_instantiations(code)),
                   configurations=list(VHDLConfiguration.find(code)),
                   references=list(VHDLReference.find(code)))

    _component_re = re.compile(
        r"[a-zA-Z]\w*\s*\:\s*(?:component)?\s*(?:(?:[a-zA-Z]\w*)\.)?([a-zA-Z]\w*)\s*"
        r"(?:generic|port) map\s*\([\s\w\=\>\,\.\)\(\+\-\'\"]*\);",
        re.IGNORECASE)

    @classmethod
    def _find_component_instantiations(cls, code):
        """
        Return the component name of all component instantiations found within the code
        """
        matches = cls._component_re.findall(code)
        return [comp_name for comp_name in matches]


class VHDLPackageBody(object):
    """
    Representation of a VHDL package body
    """
    def __init__(self, identifier):
        self.identifier = identifier

    _package_body_pattern = re.compile(r"""
        \b                             # Word boundary
        package                        # package keyword
        \s+                            # At least one whitespace
        body                           # body keyword
        \s+                            # At least one whitespace
        (?P<package>[a-zA-Z][\w]*)     # A package
        \s+                            # At least one whitespace
        is                             # is keyword
        """, re.MULTILINE | re.IGNORECASE | re.VERBOSE | re.DOTALL)

    @classmethod
    def find(cls, code):
        """
        Iterate over new instances of VHDLPackageBody for all package bodies within the code
        """
        matches = cls._package_body_pattern.finditer(code)
        for match in matches:
            yield VHDLPackageBody(match.group('package'))


class VHDLConfiguration(object):
    """
    A configuratio declaration
    """
    def __init__(self, identifier, entity):
        self.identifier = identifier
        self.entity = entity

    _configuration_re = re.compile(r"""
        \b                    # Word boundary
        configuration         # configuration keyword
        \s+                   # At least one whitespace
        (?P<id>[a-zA-Z][\w]*) # An identifier
        \s+                   # At least one whitespace
        of                    # of keyword
        \s+                   # At least one whitespace
        (?P<entity_id>[a-zA-Z][\w]*) # An identifier
        \s+                   # At least one whitespace
        is                    # is keyword
        """, re.MULTILINE | re.IGNORECASE | re.VERBOSE)

    @classmethod
    def find(cls, code):
        """
        Return tuple if library_name, unit_name for all entity instantiations found in the code
        """
        matches = cls._configuration_re.finditer(code)
        return [cls(match.group('id'), match.group('entity_id')) for match in matches]


class VHDLArchitecture(object):
    """
    Representation of a VHDL architecture
    """
    def __init__(self, identifier, entity):
        self.identifier = identifier
        self.entity = entity

    _architecture_re = re.compile(r"""
        \b                    # Word boundary
        architecture          # architecture keyword
        \s+                   # At least one whitespace
        (?P<id>[a-zA-Z][\w]*) # An identifier
        \s+                   # At least one whitespace
        of                    # of keyword
        \s+                   # At least one whitespace
        (?P<entity_id>[a-zA-Z][\w]*) # An identifier
        \s+                   # At least one whitespace
        is                    # is keyword
        """, re.MULTILINE | re.IGNORECASE | re.VERBOSE)

    @classmethod
    def find(cls, code):
        """
        Iterate over new instances of VHDLArchitecture for all architectures within the code
        """
        for arch in cls._architecture_re.finditer(code):
            identifier = arch.group('id')
            entity_id = arch.group('entity_id')
            yield VHDLArchitecture(identifier, entity_id)

PACKAGE_INSTANCE_PATTERN = (
    r'\bpackage\s+(?P<new_name>[a-zA-Z]\w*)\s+is\s+new\s+(?P<lib>[a-zA-Z]\w*)\.(?P<name>[a-zA-Z]\w*)')


class VHDLPackage(object):
    """
    Representation of a VHDL package
    """
    def __init__(self, identifier,
                 enumeration_types, record_types, array_types):
        self.identifier = identifier
        self.enumeration_types = enumeration_types
        self.record_types = record_types
        self.array_types = array_types

    _package_start_re = re.compile(r"""
        \b                    # Word boundary
        package               # package keyword
        \s+                   # At least one whitespace
        (?P<id>[a-zA-Z][\w]*) # An identifier
        \s+                   # At least one whitespace
        is                    # is keyword
        """, re.MULTILINE | re.IGNORECASE | re.VERBOSE)

    @classmethod
    def _find_normal_packages(cls, code):
        """
        Iterate over new instances of VHDLPackage for all packages within the code
        """
        for package in cls._package_start_re.finditer(code):
            identifier = package.group('id')
            package_end = re.compile(r"""
                \b                            # Word boundary
                end                           # end keyword
                (\s+package)?                 # Optional package keyword
                (\s+""" + identifier + r""")? # Optional identifier
                [\s]*                         # Potential whitespaces
                ;                             # Semicolon
                """, re.MULTILINE | re.IGNORECASE | re.VERBOSE)
            sub_code = code[package.start():]
            match = package_end.search(sub_code)
            if match:
                yield cls.parse(sub_code[:match.end()])

    _package_instance_re = re.compile("^" + PACKAGE_INSTANCE_PATTERN,
                                      re.MULTILINE | re.IGNORECASE)

    @classmethod
    def _find_package_instances(cls, code):
        """
        Find global package instances.
        Use indentation heuristic to filter out nested package instances.
        """
        references = []
        for match in cls._package_instance_re.finditer(code):
            references.append(cls(match.group("new_name"), [], [], []))
        return references

    @classmethod
    def find(cls, code):
        """
        Find normal and global generic package instances
        """
        result = list(cls._find_normal_packages(code))
        result += list(cls._find_package_instances(code))
        return result

    @classmethod
    def parse(cls, code):
        """
        Return a new VHDLPackage instance for a single package found within the code
        """
        # Extract identifier
        identifier = cls._package_start_re.match(code).group('id')
        enumeration_types = [e for e in VHDLEnumerationType.find(code)]
        record_types = [r for r in VHDLRecordType.find(code)]
        array_types = [a for a in VHDLArrayType.find(code)]

        return cls(identifier, enumeration_types, record_types, array_types)


class VHDLEntity(object):
    """
    Represents a VHDL Entity
    """
    def __init__(self, identifier, generics=None, ports=None):
        self.identifier = identifier

        if generics is not None:
            self.generics = generics
        else:
            self.generics = []

        if ports is not None:
            self.ports = ports
        else:
            self.ports = []

    def add_generic(self, identifier, subtype_code, init_value=None):
        """
        Add a generic to this entity
        """
        self.generics.append(VHDLInterfaceElement(identifier,
                                                  VHDLSubtypeIndication.parse(subtype_code),
                                                  init_value=init_value))

    def add_port(self, identifier, mode, subtype_code, init_value=None):
        """
        Add a port to this entity
        """
        self.ports.append(VHDLInterfaceElement(identifier,
                                               VHDLSubtypeIndication.parse(subtype_code),
                                               init_value=init_value,
                                               mode=mode))

    _entity_start_re = re.compile(r"""
        \b                    # Word boundary
        entity                # entity keyword
        \s+                   # At least one whitespace
        (?P<id>[a-zA-Z][\w]*) # An identifier
        \s+                   # At least one whitespace
        is                    # is keyword
        """, re.MULTILINE | re.IGNORECASE | re.VERBOSE)

    @classmethod
    def find(cls, code):
        """
        Iterates over new instances of VHDLEntity for all entities within the code
        """
        for entity in cls._entity_start_re.finditer(code):
            identifier = entity.group('id')
            sub_code = code[entity.start():]
            entity_end_re = re.compile(r"""
                \b                            # Word boundary
                end                           # end keyword
                [\s]*                         # Potential whitespaces
                (entity)?                     # Optional entity keyword
                [\s]*                         # Potential whitespaces
                (""" + identifier + r""")?    # Optional identifier
                [\s]*                         # Potential whitespaces
                ;                             # Semicolon
                """, re.MULTILINE | re.IGNORECASE | re.VERBOSE)

            match = entity_end_re.search(sub_code)
            if match:
                yield VHDLEntity.parse(sub_code[:match.end()])

    @classmethod
    def parse(cls, code):
        """
        Create a new instance by parsing the code
        """
        # Extract identifier
        re_flags = re.MULTILINE | re.IGNORECASE | re.VERBOSE
        entity_start = re.compile(r"""
            \b                    # Word boundary
            entity                # entity keyword
            \s+                   # At least one whitespace
            (?P<id>[a-zA-Z][\w]*) # An identifier
            \s+                   # At least one whitespace
            is                    # is keyword
            """, re_flags)
        identifier = entity_start.match(code).group('id')
        # Find generics and ports
        generics = cls._find_generic_clause(code)
        ports = cls._find_port_clause(code)

        return cls(identifier, generics, ports)

    @classmethod
    def _find_generic_clause(cls, code):
        """
        Find and return the generic clause code contents
        """
        re_flags = re.MULTILINE | re.IGNORECASE | re.VERBOSE
        generic_clause_start = re.compile(r"""
            \b                          # Word boundary
            generic                     # generic keyword
            [\s]*                       # Potential whitespaces
            \(                          # Opening parenthesis
            """, re_flags)
        match = generic_clause_start.search(code)
        if match:
            closing_pos = find_closing_delimiter('\\(', '\\)', code[match.end():])
            semicolon = re.compile(r"""
                [\s]*   # Potential whitespaces
                ;       # Semicolon
                """, re_flags)
            match_semicolon = semicolon.match(code[match.end() + closing_pos:])
            if match_semicolon:
                return cls._parse_generic_clause(
                    code[match.start(): match.end() + closing_pos + match_semicolon.end()])
        return []

    @classmethod
    def _find_port_clause(cls, code):
        """
        Find and return the port clause code contents
        """
        re_flags = re.MULTILINE | re.IGNORECASE | re.VERBOSE
        port_clause_start = re.compile(r"""
            ^                             # Beginning of line
            [\s]*                         # Potential whitespaces
            port                          # port keyword
            [\s]*                         # Potential whitespaces
            \(                            # Opening parenthesis
            """, re_flags)
        match = port_clause_start.search(code)
        if match:
            closing_pos = find_closing_delimiter('\\(', '\\)', code[match.end():])
            semicolon = re.compile(r"""
                [\s]*   # Potential whitespaces
                ;       # Semicolon
                """, re_flags)
            match_semicolon = semicolon.match(code[match.end() + closing_pos:])
            if match_semicolon:
                return cls._parse_port_clause(
                    code[match.start(): match.end() + closing_pos + match_semicolon.end()])
        return []

    @staticmethod
    def _split_not_in_par(string, sep):
        """
        Split string at all occurences of sep but not inside of a parenthesis
        """
        result = []
        count = 0
        split = []
        for char in string:
            if char == '(':
                count += 1
            elif char == ')':
                count -= 1

            if char == sep and count == 0:
                result.append("".join(split))
                split = []
            else:
                split.append(char)

        if split:
            result.append("".join(split))

        return result

    _package_generic_re = re.compile(r"\s*package\s+", re.MULTILINE | re.IGNORECASE)
    _type_generic_re = re.compile(r"\s*type\s+", re.MULTILINE | re.IGNORECASE)
    _function_generic_re = re.compile(r"\s*(impure\s+)?(function|procedure)\s+", re.MULTILINE | re.IGNORECASE)

    @classmethod
    def _parse_generic_clause(cls, code):
        """
        Parse the generic clause and return a list of interface elements
        """
        # The generic list is between the outer parenthesis
        generic_list_string = code[code.find('(') + 1: code.rfind(')')]

        # Split the interface elements
        interface_elements = cls._split_not_in_par(generic_list_string, ';')

        generic_list = []
        # Add interface elements to the generic list
        for interface_element in interface_elements:

            if cls._package_generic_re.match(interface_element) is not None:
                # Ignore package generics
                continue

            if cls._type_generic_re.match(interface_element) is not None:
                # Ignore type generics
                continue

            if cls._function_generic_re.match(interface_element) is not None:
                # Ignore function generics
                continue

            generic_list.append(VHDLInterfaceElement.parse(interface_element))

        return generic_list

    @classmethod
    def _parse_port_clause(cls, code):
        """
        Parse the port clause and return a list of interface elements
        """
        # The port list is between the outer parenthesis
        port_list_string = code[code.find('(') + 1: code.rfind(')')]

        # Split the interface elements
        interface_elements = port_list_string.split(';')

        port_list = []
        # Add interface elements to the port list
        for interface_element in interface_elements:
            port_list.append(VHDLInterfaceElement.parse(interface_element, is_signal=True))

        return port_list


class VHDLContext(object):
    """
    Represents a VHDL 2008 context
    """
    def __init__(self, identifier):
        self.identifier = identifier

    _context_start_re = re.compile(r"""
        \b                    # Word boundary
        context               # context keyword
        \s+                   # At least one whitespace
        (?P<id>[a-zA-Z][\w]*) # An identifier
        \s+                   # At least one whitespace
        is                    # is keyword
        """, re.MULTILINE | re.IGNORECASE | re.VERBOSE)

    @classmethod
    def find(cls, code):
        """
        Iterate over new instances of VHDLContext for a contexts found in the code
        """
        for context in cls._context_start_re.finditer(code):
            identifier = context.group('id')
            yield VHDLContext(identifier=identifier)


class VHDLSubtypeIndication(object):
    """
    Represents a VHDL subtype indication
    """
    def __init__(self, code, type_mark, constraint, array_type):
        self.code = code
        self.type_mark = type_mark
        self.constraint = constraint
        self.array_type = array_type

    @classmethod
    def parse(cls, code):
        """
        Returns a new instance from parsing the code
        """
        # Extract type mark and find out if it's an array type and if a constraint is given.
        re_flags = re.MULTILINE | re.IGNORECASE | re.VERBOSE
        subtype_indication_start = re.compile(r"""
            ^                             # Beginning of line
            [\s]*                         # Potential whitespaces
            (?P<type_mark>[a-zA-Z][\w]*)   # An type mark
            [\s]*                         # Potential whitespaces
            (?P<constraint>\(.*\))?
            """, re_flags)
        subtype_indication_declaration = subtype_indication_start.match(code)
        type_mark = subtype_indication_declaration.group('type_mark')
        constraint = subtype_indication_declaration.group('constraint')

        array_type = type_mark == 'std_logic_vector'
        return cls(code, type_mark, constraint, array_type)

    def __str__(self):
        return self.code


class VHDLInterfaceElement(object):
    """
    Represents a VHDL interface element
    """
    def __init__(self, identifier, subtype_indication, mode=None, init_value=None):
        self.identifier = identifier
        self.mode = mode
        self.subtype_indication = subtype_indication
        self.init_value = init_value

    def without_mode(self):
        """
        @returns A copy of this interface element without a mode
        """
        return VHDLInterfaceElement(self.identifier,
                                    self.subtype_indication,
                                    init_value=self.init_value)

    @classmethod
    def parse(cls, code, is_signal=False):
        """
        Returns a new instance by parsing the code
        """
        if is_signal:
            # Remove 'signal' string if a signal is beeing parsed
            code = code.replace("signal", "")

        interface_element_string = code

        # Extract the identifier
        identifier = interface_element_string.split(':')[0].strip()

        # Extract subtype indication and mode (if any)

        mode_split = interface_element_string.split(':')[1].strip().split(None, 1)
        if cls._is_mode(mode_split[0]):
            mode = mode_split[0]
            subtype_indication = VHDLSubtypeIndication.parse(mode_split[1])
        else:
            mode = None
            subtype_indication = VHDLSubtypeIndication.parse(interface_element_string.split(':')[1].strip())

        # Extract initial value
        init_value_split = interface_element_string.split(':=')
        if len(init_value_split) > 1:
            init_value = init_value_split[1].strip()
        else:
            init_value = None

        return cls(identifier, subtype_indication, mode, init_value)

    @staticmethod
    def _is_mode(code):
        """
        Return True if the code is a mode keyword
        """
        return code in ('in', 'out', 'inout', 'buffer', 'linkage')

    def __str__(self):
        code = self.identifier + " : "

        if self.mode is not None:
            code += self.mode + " "

        code += str(self.subtype_indication)

        if self.init_value is not None:
            code += " := " + self.init_value

        return code


class VHDLEnumerationType(object):
    """Represents a VHDL enumeration type"""
    def __init__(self, identifier, literals):
        self.identifier = identifier
        self.literals = literals

    _enum_declaration_re = re.compile(r"""
        \b                    # Word boundary
        type
        \s+
        (?P<id>[a-zA-Z][\w]*)       # An identifier
        \s+
        is
        \s*\(\s*
        (?P<literals>[a-zA-Z][\w]* # First enumeration literal
        (\s*,\s*[a-zA-Z][\w]*)*)   # More enumeration literals
        \s*\)\s*;""", re.MULTILINE | re.IGNORECASE | re.VERBOSE)

    @classmethod
    def find(cls, code):
        """
        Find enumeration types in the code
        """
        for enum_type in cls._enum_declaration_re.finditer(code):
            identifier = enum_type.group('id')
            literals = [e.strip() for e in enum_type.group('literals').split(',')]
            yield cls(identifier, literals)


class VHDLElementDeclaration(object):
    """Represents a VHDL element declaration"""
    def __init__(self, identifier_list, subtype_indication):
        self.identifier_list = identifier_list
        self.subtype_indication = subtype_indication


class VHDLRecordType(object):
    """Represents a VHDL record type"""
    def __init__(self, identifier, elements):
        self.identifier = identifier
        self.elements = elements

    _record_declaration_re = re.compile(r"""
        \b                    # Word boundary
        type
        \s+
        (?P<id>[a-zA-Z][\w]*)       # An identifier
        \s+
        is
        \s+
        record
        (?P<elements>.*?)end\s+record""", re.MULTILINE | re.IGNORECASE | re.VERBOSE | re.DOTALL)

    @classmethod
    def find(cls, code):
        """
        Find all record types in the code
        """
        for record_type in cls._record_declaration_re.finditer(code):
            identifier = record_type.group('id')
            elements = record_type.group('elements').split(';')
            parsed_elements = []
            for element in elements:
                if ':' in element:
                    identifier_list_and_subtype_indication = element.split(':')
                    identifier_list = [i.strip() for i in identifier_list_and_subtype_indication[0].split(',')]
                    subtype_indication = VHDLSubtypeIndication.parse(identifier_list_and_subtype_indication[1].strip())
                    parsed_elements.append(VHDLElementDeclaration(identifier_list, subtype_indication))
            yield cls(identifier, parsed_elements)


class VHDLRange(object):
    """Represents a VHDL Range"""
    def __init__(self, range_type=None, left=None, right=None, attribute=None):
        self.range_type = range_type
        self.left = left
        self.right = right
        self.attribute = attribute


class VHDLArrayType(object):
    """Represents a VHDL array type"""
    def __init__(self, identifier, subtype_indication, range1, range2):
        self.identifier = identifier
        self.subtype_indication = subtype_indication
        self.range1 = range1
        self.range2 = range2

    _constrained_ranges_re = re.compile(r"""
        \s*(?P<range_left1>.+?)
        \s+(to|downto)\s+
        (?P<range_right1>.+?)\s*
        (,
        \s*(?P<range_left2>.+?)
        \s+(to|downto)\s+
        (?P<range_right2>.+?)\s*)?""", re.MULTILINE | re.IGNORECASE | re.VERBOSE | re.DOTALL)

    _range_attribute_ranges_re = re.compile(r"""
        \s*(?P<range_attribute>[a-zA-Z][\w]*'range)\s*""", re.MULTILINE | re.IGNORECASE | re.VERBOSE | re.DOTALL)

    _unconstrained_ranges_re = re.compile(r"""
        \s*(?P<range_type1>[a-zA-Z][\w]*)
        \s+range\s+<>\s*
        (,
        \s*(?P<range_type2>[a-zA-Z][\w]*)
        \s+range\s+<>\s*)?""", re.MULTILINE | re.IGNORECASE | re.VERBOSE | re.DOTALL)

    _constrained_range_re = re.compile(r"""
        \s*(?P<range_left>.+?)
        \s+(to|downto)\s+
        (?P<range_right>.+?)\s*""", re.MULTILINE | re.IGNORECASE | re.VERBOSE | re.DOTALL)

    _range_attribute_range_re = re.compile(r"""
        \s*(?P<range_attribute>[a-zA-Z][\w]*'range)\s*""", re.MULTILINE | re.IGNORECASE | re.VERBOSE | re.DOTALL)

    _unconstrained_range_re = re.compile(r"""
        \s*(?P<range_type>[a-zA-Z][\w]*)
        \s+range\s+<>\s*""", re.MULTILINE | re.IGNORECASE | re.VERBOSE | re.DOTALL)

    _array_declaration_re = re.compile(r"""
        \b                    # Word boundary
        type
        \s+
        (?P<id>[a-zA-Z][\w]*)
        \s+
        is
        \s+
        array
        \s*\(
        (?P<ranges>.*?)
        \)\s+of\s+
        (?P<subtype_indication>.*?)\s*;""", re.MULTILINE | re.IGNORECASE | re.VERBOSE | re.DOTALL)

    @classmethod
    def find(cls, code):
        """Iterate over new instances of VHDLArrayType for all array types within the code"""
        for array_type in cls._array_declaration_re.finditer(code):
            identifier = array_type.group('id')
            subtype_indication = VHDLSubtypeIndication.parse(array_type.group('subtype_indication'))
            ranges = array_type.group('ranges')
            range1_str, range2_str = cls._split_ranges(ranges)
            range1 = cls._parse_range(range1_str)
            range2 = cls._parse_range(range2_str)

            yield cls(identifier, subtype_indication, range1, range2)

    @staticmethod
    def _split_ranges(ranges):
        """Splits 2D ranges in two. 1D ranges will return None as the second range"""
        level = 0
        index = 0
        if ',' in ranges:
            for char in ranges:
                if char == ',' and level == 0:
                    return ranges[:index], ranges[index + 1:]
                elif char == '(':
                    level += 1
                elif char == ')':
                    level -= 1
                index += 1

        return ranges, None

    @classmethod
    def _parse_range(cls, the_range):
        """Extracts range type, left and right boundary as well as the range when the 'range attribute
        is used"""
        if the_range is None:
            return VHDLRange()

        unconstrained_range = cls._unconstrained_range_re.match(the_range)
        if unconstrained_range is not None:
            range_type = unconstrained_range.group('range_type')
            return VHDLRange(range_type)
        else:
            constrained_range = cls._constrained_range_re.match(the_range)
            range_attribute = cls._range_attribute_range_re.match(the_range)
            if constrained_range is not None:
                range_left = constrained_range.group('range_left')
                range_right = constrained_range.group('range_right')
                return VHDLRange(None, range_left, range_right)
            elif range_attribute is not None:
                range_attribute = range_attribute.group('range_attribute')
                return VHDLRange(attribute=range_attribute)

        return VHDLRange()


def find_closing_delimiter(start, end, code):
    """
    Find the balanced closing position within the code.

    The balanced closing position is defined as the first position of an end marker
    where the number of previous start and end markers are equal
    """
    delimiter_pattern = start + '|' + end
    start = start.replace('\\', '')
    end = end.replace('\\', '')
    delimiters = re.compile(delimiter_pattern)
    count = 1
    for delimiter in delimiters.finditer(code):
        if delimiter.group() == start:
            count += 1
        else:
            count -= 1

        if count == 0:
            return delimiter.end()
    raise ValueError('Failed to find closing delimiter to ' + start + ' in ' + code + '.')


class VHDLReference(object):
    """
    Reference to design unit
    """
    _reference_types = ("package",
                        "context",
                        "entity",
                        "configuration")

    _uses_re = re.compile(r"""
            \b                             # Word boundary
            (?P<use_type>use|context)      # use or context keyword
            \s+                            # At least one whitespace
            (?P<id>[a-zA-Z][\w]*(\.[a-zA-Z][\w]*){1,2})
            (?P<extra>(\s*,\s*[a-zA-Z][\w]*(\.[a-zA-Z][\w]*){1,2})*)
            \s*                            # Potential whitespaces
            ;                              # Semi-colon
    """, re.MULTILINE | re.IGNORECASE | re.VERBOSE)

    @classmethod
    def _find_uses(cls, code):
        """
        Find all the libraries and use clasues within the code
        """

        def get_ids(match):
            """
            Get all ids found within the match taking the optinal extra ids of
            library and use clauses into account such as:

            use foo, bar;

            or

            library foo, bar;
            """
            ids = [match.group('id').strip()]
            if match.group('extra'):
                ids += [name.strip() for name in match.group('extra').split(',')[1:]]
            return ids

        references = []
        for match in cls._uses_re.finditer(code):
            for uses in get_ids(match):
                uses = uses.split(".")

                names_within = uses[2:] if len(uses) > 2 else (None,)
                for name_within in names_within:
                    ref = cls(reference_type="package" if match.group("use_type") == "use" else "context",
                              library=uses[0],
                              design_unit=uses[1],
                              name_within=name_within)

                    references.append(ref)
        return references

    _entity_reference_re = re.compile(
        r'\bentity\s+(?P<lib>[a-zA-Z]\w*)\.(?P<ent>[a-zA-Z]\w*)\s*(\((?P<arch>[a-zA-Z]\w*)\))?',
        re.MULTILINE | re.IGNORECASE)

    @classmethod
    def _find_entity_references(cls, code):
        """
        Find all entity references from instantiations or block configurations
        """
        references = []
        for match in cls._entity_reference_re.finditer(code):
            if match.group("arch") is None:
                references.append(cls('entity', match.group("lib"), match.group("ent")))
            else:
                references.append(cls('entity', match.group("lib"), match.group("ent"), match.group("arch")))
        return references

    _configuration_reference_re = re.compile(
        r'\bconfiguration\s+(?P<lib>[a-zA-Z]\w*)\.(?P<cfg>[a-zA-Z]\w*)',
        re.MULTILINE | re.IGNORECASE)

    @classmethod
    def _find_configuration_references(cls, code):
        """
        Find all configuration references within block configurations
        """
        references = []
        for match in cls._configuration_reference_re.finditer(code):
            references.append(cls('configuration', match.group("lib"), match.group("cfg")))
        return references

    _package_instance_re = re.compile(PACKAGE_INSTANCE_PATTERN,
                                      re.MULTILINE | re.IGNORECASE)

    @classmethod
    def _find_package_instance_references(cls, code):
        """
        Finds all reference causes by package instantiation
        """
        references = []
        for match in cls._package_instance_re.finditer(code):
            references.append(cls('package', match.group("lib"), match.group("name")))
        return references

    @classmethod
    def find(cls, code):
        """
        Find entity, use, context and configuration references within the code
        """
        return (cls._find_uses(code) +
                cls._find_entity_references(code) +
                cls._find_configuration_references(code) +
                cls._find_package_instance_references(code))

    def __init__(self, reference_type, library, design_unit, name_within=None):
        assert reference_type in self._reference_types
        self.reference_type = reference_type
        self.library = library
        self.design_unit = design_unit

        # String "all" may be used to denote all names within
        self.name_within = name_within

    def __repr__(self):
        return "VHDLReference(%r, %r, %r, %r)" % (
            self.reference_type,
            self.library,
            self.design_unit,
            self.name_within)

    def __eq__(self, other):
        return (self.reference_type == other.reference_type and
                self.library == other.library and
                self.design_unit == other.design_unit and
                self.name_within == other.name_within)

    def copy(self):
        return VHDLReference(self.reference_type,
                             self.library,
                             self.design_unit,
                             self.name_within)

    def is_entity_reference(self):
        return self.reference_type == 'entity'

    def is_package_reference(self):
        return self.reference_type == 'package'

    def reference_all_names_within(self):
        return self.name_within == "all"


VHDL_REMOVE_COMMENT_RE = re.compile(r'--[^\n]*')


def remove_comments(code):
    """
    Return the code with comments removed
    """
    return VHDL_REMOVE_COMMENT_RE.sub('', code)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2015-2017, Lars Asplund lars.anders.asplund@gmail.com

"""
Create simulator instances
"""

from os.path import exists
import os
from vunit.modelsim_interface import ModelSimInterface
from vunit.activehdl_interface import ActiveHDLInterface
from vunit.rivierapro_interface import RivieraProInterface
from vunit.ghdl_interface import GHDLInterface
from vunit.incisive_interface import IncisiveInterface
from vunit.simulator_interface import (BooleanOption,
                                       ListOfStringOption,
                                       VHDLAssertLevelOption)


class SimulatorFactory(object):
    """
    Create simulator instances
    """

    @staticmethod
    def supported_simulators():
        """
        Return a list of supported simulator classes
        """
        return [ModelSimInterface,
                RivieraProInterface,
                ActiveHDLInterface,
                GHDLInterface,
                IncisiveInterface]

    def _extract_compile_options(self):
        """
        Return all supported compile options
        """
        result = {}
        for sim_class in self.supported_simulators():
            for opt in sim_class.compile_options:
                assert hasattr(opt, "name")
                assert hasattr(opt, "validate")
                assert opt.name.startswith(sim_class.name + ".")
                assert opt.name not in result
                result[opt.name] = opt
        return result

    def _extract_sim_options(self):
        """
        Return all supported sim options
        """
        result = dict((opt.name, opt) for opt in
                      [VHDLAssertLevelOption(),
                       BooleanOption("disable_ieee_warnings"),
                       ListOfStringOption("pli")])

        for sim_class in self.supported_simulators():
            for opt in sim_class.sim_options:
                assert hasattr(opt, "name")
                assert hasattr(opt, "validate")
                assert opt.name.startswith(sim_class.name + ".")
                assert opt.name not in result
                result[opt.name] = opt

        return result

    def check_sim_option(self, name, value):
        """
        Check that sim_option has legal name and value
        """
        known_options = sorted(list(self._sim_options.keys()))

        if name not in self._sim_options:
            raise ValueError("Unknown sim_option %r, expected one of %r" %
                             (name, known_options))

        self._sim_options[name].validate(value)

    def check_compile_option_name(self, name):
        """
        Check that the compile option is valid
        """
        known_options = sorted(list(self._compile_options.keys()))
        if name not in known_options:
            raise ValueError("Unknown compile_option %r, expected one of %r" %
                             (name, known_options))

    def check_compile_option(self, name, value):
        """
        Check that the compile option is valid
        """
        self.check_compile_option_name(name)
        self._compile_options[name].validate(value)

    def _select_simulator(self):
        """
        Select simulator class, either from VUNIT_SIMULATOR environment variable
        or the first available
        """
        environ_name = "VUNIT_SIMULATOR"

        available_simulators = self.available_simulators()
        name_mapping = {simulator_class.name: simulator_class for simulator_class in self.supported_simulators()}
        if not available_simulators:
            return None

        if environ_name in os.environ:
            simulator_name = os.environ[environ_name]
            if simulator_name not in name_mapping:
                raise RuntimeError(
                    ("Simulator from " + environ_name + " environment variable %r is not supported. "
                     "Supported simulators are %r")
                    % (simulator_name, name_mapping.keys()))
            simulator_class = name_mapping[simulator_name]
        else:
            simulator_class = available_simulators[0]

        return simulator_class

    def add_arguments(self, parser, for_all_simulators=False):
        """
        Add command line arguments to parser
        """

        if for_all_simulators or (self.has_simulator and self._simulator_class.supports_gui_flag):
            parser.add_argument('-g', '--gui',
                                action="store_true",
                                default=False,
                                help=("Open test case(s) in simulator gui with top level pre loaded"))

        if for_all_simulators:
            for sim in self.supported_simulators():
                sim.add_arguments(parser)
        elif self._simulator_class is not None:
            self._simulator_class.add_arguments(parser)

    def __init__(self):
        self._available_simulators = [simulator_class
                                      for simulator_class in self.supported_simulators()
                                      if simulator_class.is_available()]
        self._simulator_class = self._select_simulator()
        self._compile_options = self._extract_compile_options()
        self._sim_options = self._extract_sim_options()

    def available_simulators(self):
        """
        Return a list of available simulators
        """
        return self._available_simulators

    def package_users_depend_on_bodies(self):
        """
        Returns True when package users also depend on package bodies
        """
        if self._simulator_class is not None:
            return self._simulator_class.package_users_depend_on_bodies

        return False

    def supports_vhdl_2008_contexts(self):
        """
        Returns True when this simulator supports VHDL 2008 contexts
        """
        if self._simulator_class is not None:
            return self._simulator_class.supports_vhdl_2008_contexts()

        return True

    def get_osvvm_coverage_api(self):
        """
        Returns simulator name when OSVVM coverage API is supported, None otherwise.
        """
        if self._simulator_class is not None:
            return self._simulator_class.get_osvvm_coverage_api()

        return None

    def supports_vhdl_package_generics(self):
        """
        Returns True when this simulator supports VHDL package generics
        """
        if self._simulator_class is not None:
            return self._simulator_class.supports_vhdl_package_generics()

        return False

    @property
    def has_simulator(self):
        return self._simulator_class is not None

    @property
    def simulator_name(self):
        """
        Return the name of the selected simulator or none
        """
        if self._simulator_class is None:
            return "none"

        return self._simulator_class.name

    def create(self, args, simulator_output_path):
        """
        Create new simulator instance
        """

        if not self.has_simulator:
            raise RuntimeError("No available simulator detected. "
                               "Simulator executables must be available in PATH environment variable.")

        if not exists(simulator_output_path):
            os.makedirs(simulator_output_path)

        simulator_if = self._simulator_class.from_args(simulator_output_path, args)
        simulator_if.set_output_path(simulator_output_path)
        return simulator_if


SIMULATOR_FACTORY = SimulatorFactory()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2015-2016, Lars Asplund lars.anders.asplund@gmail.com

"""
Verify that all external run scripts work correctly
"""


import unittest
from os import environ
from os.path import join, dirname
from subprocess import call
import sys
from vunit import ROOT
from vunit.builtins import VHDL_PATH
from vunit.test.common import has_simulator, check_report, simulator_is


def simulator_supports_verilog():
    """
    Returns True if simulator supports Verilog
    """
    return simulator_is("modelsim", "incisive")


# pylint: disable=too-many-public-methods
@unittest.skipUnless(has_simulator(), "Requires simulator")
class TestExternalRunScripts(unittest.TestCase):
    """
    Verify that example projects run correctly
    """

    def test_vhdl_uart_example_project(self):
        self.check(join(ROOT, "examples", "vhdl", "uart", "run.py"))

    @unittest.skipUnless(simulator_supports_verilog(), "Verilog")
    def test_verilog_uart_example_project(self):
        self.check(join(ROOT, "examples", "verilog", "uart", "run.py"))

    @unittest.skipUnless(simulator_supports_verilog(), "Verilog")
    def test_verilog_ams_example(self):
        self.check(join(ROOT, "examples", "verilog", "verilog_ams", "run.py"))
        check_report(self.report_file,
                     [("passed", "lib.tb_dut.Test that pass"),
                      ("failed", "lib.tb_dut.Test that fail")])

    def test_vhdl_logging_example_project(self):
        self.check(join(ROOT, "examples", "vhdl", "logging", "compile.py"), args=["--compile"])

    def test_vhdl_check_example_project(self):
        self.check(join(ROOT, "examples", "vhdl", "check", "run.py"))

    def test_vhdl_generate_tests_example_project(self):
        self.check(join(ROOT, "examples", "vhdl", "generate_tests", "run.py"))
        check_report(self.report_file,
                     [("passed", "lib.tb_generated.data_width=1,sign=False.Test 1"),
                      ("passed", "lib.tb_generated.data_width=1,sign=True.Test 1"),
                      ("passed", "lib.tb_generated.data_width=2,sign=False.Test 1"),
                      ("passed", "lib.tb_generated.data_width=2,sign=True.Test 1"),
                      ("passed", "lib.tb_generated.data_width=3,sign=False.Test 1"),
                      ("passed", "lib.tb_generated.data_width=3,sign=True.Test 1"),
                      ("passed", "lib.tb_generated.data_width=4,sign=False.Test 1"),
                      ("passed", "lib.tb_generated.data_width=4,sign=True.Test 1"),
                      ("passed", "lib.tb_generated.data_width=16,sign=True.Test 2")])

    def test_vhdl_composite_generics_example_project(self):
        self.check(join(ROOT, "examples", "vhdl", "composite_generics", "run.py"))
        check_report(self.report_file,
                     [("passed", "tb_lib.tb_composite_generics.VGA.Test 1"),
                      ("passed", "tb_lib.tb_composite_generics.tiny.Test 1")])

    def test_vhdl_array_example_project(self):
        self.check(join(ROOT, "examples", "vhdl", "array", "run.py"))

    def test_vhdl_user_guide_example_project(self):
        self.check(join(ROOT, "examples", "vhdl", "user_guide", "run.py"), exit_code=1)
        check_report(self.report_file,
                     [("passed", "lib.tb_example.all"),
                      ("passed", "lib.tb_example_many.test_pass"),
                      ("failed", "lib.tb_example_many.test_fail")])

    @unittest.skipUnless(simulator_supports_verilog(), "Verilog")
    def test_verilog_user_guide_example_project(self):
        self.check(join(ROOT, "examples", "verilog", "user_guide", "run.py"), exit_code=1)
        check_report(self.report_file,
                     [("passed", "lib.tb_example.Test that a successful test case passes"),
                      ("failed", "lib.tb_example.Test that a failing test case actually fails"),
                      ("failed", "lib.tb_example.Test that a test case that takes too long time fails with a timeout")])

    def test_vhdl_osvvm_integration_example_project(self):
        self.check(join(ROOT, "examples", "vhdl", "osvvm_integration", "run.py"), exit_code=1)
        check_report(self.report_file,
                     [("passed", "lib.tb_alertlog_demo_global_with_comments.Test passing alerts"),
                      ("passed", "lib.tb_alertlog_demo_hierarchy_with_comments.Test passing alerts"),
                      ("passed", "lib.tb_alertlog_demo_global.Test passing alerts"),
                      ("passed", "lib.tb_alertlog_demo_hierarchy.Test passing alerts"),
                      ("failed", "lib.tb_alertlog_demo_global_with_comments.Test failing alerts"),
                      ("failed", "lib.tb_alertlog_demo_hierarchy_with_comments.Test failing alerts"),
                      ("failed", "lib.tb_alertlog_demo_global.Test failing alerts"),
                      ("failed", "lib.tb_alertlog_demo_hierarchy.Test failing alerts")])

    def test_vhdl_com_example_project(self):
        self.check(join(ROOT, "examples", "vhdl", "com", "run.py"))

    def test_array_vhdl_2008(self):
        self.check(join(VHDL_PATH, "array", "run.py"))

    def test_check_vhdl_2008(self):
        self.check(join(VHDL_PATH, "check", "run.py"))

    def test_check_vhdl_2002(self):
        self.check(join(VHDL_PATH, "check", "run.py"),
                   vhdl_standard='2002')

    def test_check_vhdl_93(self):
        self.check(join(VHDL_PATH, "check", "run.py"),
                   vhdl_standard='93')

    def test_logging_vhdl_2008(self):
        self.check(join(VHDL_PATH, "logging", "run.py"))

    def test_logging_vhdl_2002(self):
        self.check(join(VHDL_PATH, "logging", "run.py"),
                   vhdl_standard='2002')

    def test_logging_vhdl_93(self):
        self.check(join(VHDL_PATH, "logging", "run.py"),
                   vhdl_standard='93')

    def test_run_vhdl_2008(self):
        self.check(join(VHDL_PATH, "run", "run.py"))

    def test_run_vhdl_2002(self):
        self.check(join(VHDL_PATH, "run", "run.py"),
                   vhdl_standard='2002')

    def test_run_vhdl_93(self):
        self.check(join(VHDL_PATH, "run", "run.py"),
                   vhdl_standard='93')

    def test_string_ops_vhdl_2008(self):
        self.check(join(VHDL_PATH, "string_ops", "run.py"))

    def test_string_ops_vhdl_2002(self):
        self.check(join(VHDL_PATH, "string_ops", "run.py"),
                   vhdl_standard='2002')

    def test_string_ops_vhdl_93(self):
        self.check(join(VHDL_PATH, "string_ops", "run.py"),
                   vhdl_standard='93')

    def test_dictionary_vhdl_2008(self):
        self.check(join(VHDL_PATH, "dictionary", "run.py"))

    def test_dictionary_vhdl_2002(self):
        self.check(join(VHDL_PATH, "dictionary", "run.py"),
                   vhdl_standard='2002')

    def test_dictionary_vhdl_93(self):
        self.check(join(VHDL_PATH, "dictionary", "run.py"),
                   vhdl_standard='93')

    def test_path_vhdl_2008(self):
        self.check(join(VHDL_PATH, "path", "run.py"))

    def test_path_vhdl_2002(self):
        self.check(join(VHDL_PATH, "path", "run.py"),
                   vhdl_standard='2002')

    def test_path_vhdl_93(self):
        self.check(join(VHDL_PATH, "path", "run.py"),
                   vhdl_standard='93')

    def test_com_vhdl_2008(self):
        self.check(join(VHDL_PATH, "com", "run.py"))

    def test_com_debug_vhdl_2008(self):
        self.check(join(VHDL_PATH, "com", "run.py"),
                   args=["--use-debug-codecs"])

    def setUp(self):
        self.output_path = join(dirname(__file__), "external_run_out")
        self.report_file = join(self.output_path, "xunit.xml")

    def check(self, run_file, args=None, vhdl_standard='2008', exit_code=0):
        """
        Run external run file and verify exit code
        """
        args = args if args is not None else []
        new_env = environ.copy()
        new_env["VUNIT_VHDL_STANDARD"] = vhdl_standard
        retcode = call([sys.executable, run_file,
                        "--clean",
                        "--output-path=%s" % self.output_path,
                        "--xunit-xml=%s" % self.report_file] + args,
                       env=new_env)
        self.assertEqual(retcode, exit_code)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2016, Lars Asplund lars.anders.asplund@gmail.com

# pylint: disable=too-many-public-methods

"""
Tests the test test_bench module
"""


import unittest
from os.path import join, dirname, exists
from shutil import rmtree

from vunit.test_bench import TestBench
from vunit.ostools import renew_path, write_file
from vunit.test.mock_2or3 import mock


class TestTestBench(unittest.TestCase):
    """
    Tests the test bench
    """

    def setUp(self):
        self.simulator_if = 'simulator_if'
        self.output_path = out()
        renew_path(self.output_path)

    def tearDown(self):
        if exists(self.output_path):
            rmtree(self.output_path)

    def test_that_single_vhdl_test_is_created(self):
        design_unit = Entity('tb_entity')
        test_bench = TestBench(design_unit)
        tests = self.create_tests(test_bench)
        self.assert_has_tests(tests, ["lib.tb_entity.all"])

    @staticmethod
    def test_no_architecture_at_creation():
        design_unit = Entity('tb_entity', no_arch=True)
        TestBench(design_unit)

    def test_no_architecture_gives_runtime_error(self):
        design_unit = Entity('tb_entity', no_arch=True)
        test_bench = TestBench(design_unit)
        try:
            self.create_tests(test_bench)
        except RuntimeError as exc:
            self.assertEqual(str(exc),
                             "Test bench 'tb_entity' has no architecture.")
        else:
            assert False, "RuntimeError not raised"

    def test_that_single_verilog_test_is_created(self):
        design_unit = Module('tb_module')
        test_bench = TestBench(design_unit)
        tests = self.create_tests(test_bench)
        self.assert_has_tests(tests, ["lib.tb_module.all"])

    def test_create_default_test(self):
        design_unit = Entity('tb_entity')
        design_unit.generic_names = ["runner_cfg"]
        test_bench = TestBench(design_unit)
        tests = self.create_tests(test_bench)
        self.assert_has_tests(tests, ["lib.tb_entity.all"])

    def test_multiple_architectures_are_not_allowed_for_test_bench(self):
        design_unit = Entity('tb_entity')
        design_unit.generic_names = ["runner_cfg"]
        design_unit.add_architecture('arch2')
        try:
            TestBench(design_unit)
        except RuntimeError as exc:
            self.assertEqual(str(exc),
                             "Test bench not allowed to have multiple architectures. "
                             "Entity tb_entity has arch:file.vhd, arch2:arch2.vhd")
        else:
            assert False, "RuntimeError not raised"

    def test_creates_tests_vhdl(self):
        design_unit = Entity('tb_entity', contents='''\
if run("Test 1")
--if run("Test 2")
if run("Test 3")
if run("Test 4")
if run("Test 5") or run("Test 6")
if run  ("Test 7")
if run( "Test 8"  )
if ((run("Test 9")))
if my_protected_variable.run("Test 10")
''')
        design_unit.generic_names = ["runner_cfg"]
        test_bench = TestBench(design_unit)
        tests = self.create_tests(test_bench)
        self.assert_has_tests(tests,
                              ["lib.tb_entity.Test 1",
                               "lib.tb_entity.Test 3",
                               "lib.tb_entity.Test 4",
                               "lib.tb_entity.Test 5",
                               "lib.tb_entity.Test 6",
                               "lib.tb_entity.Test 7",
                               "lib.tb_entity.Test 8",
                               "lib.tb_entity.Test 9"])

    def test_creates_tests_verilog(self):
        design_unit = Module('tb_module', contents='''\
`TEST_CASE("Test 1")
`TEST_CASE  ("Test 2")
`TEST_CASE( "Test 3"  )
// `TEST_CASE("Test 4")
/*
 `TEST_CASE("Test 5")
*/
/* `TEST_CASE("Test 6") */
`TEST_CASE("Test 7")
/* comment */
`TEST_CASE("Test 8")
/* comment */
''')
        design_unit.generic_names = ["runner_cfg"]
        test_bench = TestBench(design_unit)
        tests = self.create_tests(test_bench)
        self.assert_has_tests(tests,
                              ["lib.tb_module.Test 1",
                               "lib.tb_module.Test 2",
                               "lib.tb_module.Test 3",
                               "lib.tb_module.Test 7",
                               "lib.tb_module.Test 8"])

    @mock.patch("vunit.test_bench.LOGGER")
    def test_duplicate_tests_cause_error(self, mock_logger):
        design_unit = Entity('tb_entity', contents='''\
if run("Test_1")
--if run("Test_1")
if run("Test_3")
if run("Test_2")
if run("Test_3")
if run("Test_3")
if run("Test_2")
''')
        design_unit.generic_names = ["runner_cfg"]
        self.assertRaises(RuntimeError, TestBench, design_unit)

        error_calls = mock_logger.error.call_args_list
        self.assertEqual(len(error_calls), 2)
        call0_args = error_calls[0][0]
        self.assertIn("Test_3", call0_args)
        self.assertIn(design_unit.file_name, call0_args)

        call1_args = error_calls[1][0]
        self.assertIn("Test_2", call1_args)
        self.assertIn(design_unit.file_name, call1_args)

    def test_keyerror_on_non_existent_test(self):
        design_unit = Entity('tb_entity', contents='if run("Test")')
        design_unit.generic_names = ["runner_cfg", "name"]
        test_bench = TestBench(design_unit)
        self.assertRaises(KeyError, test_bench.get_test_case, "Non existent test")

    def test_creates_tests_when_adding_architecture_late(self):
        design_unit = Entity('tb_entity', no_arch=True)
        design_unit.generic_names = ["runner_cfg"]
        test_bench = TestBench(design_unit)

        design_unit.add_architecture("arch", contents='''\
if run("Test_1")
--if run("Test_2")
if run("Test_3")
''')
        tests = self.create_tests(test_bench)
        self.assert_has_tests(tests,
                              ["lib.tb_entity.Test_1",
                               "lib.tb_entity.Test_3"])

    def test_scan_tests_from_file(self):
        design_unit = Entity('tb_entity')
        design_unit.generic_names = ["runner_cfg"]
        test_bench = TestBench(design_unit)

        file_name = out("file.vhd")
        write_file(file_name, '''\
if run("Test_1")
if run("Test_2")
''')
        test_bench.scan_tests_from_file(file_name)
        tests = self.create_tests(test_bench)
        self.assert_has_tests(tests,
                              ["lib.tb_entity.Test_1",
                               "lib.tb_entity.Test_2"])

    def test_scan_tests_from_missing_file(self):
        design_unit = Entity('tb_entity')
        design_unit.generic_names = ["runner_cfg"]
        test_bench = TestBench(design_unit)

        try:
            test_bench.scan_tests_from_file("missing.vhd")
        except ValueError as exc:
            self.assertEqual(str(exc),
                             "File 'missing.vhd' does not exist")
        else:
            assert False, "ValueError not raised"

    def test_does_not_add_all_suffix_with_named_configurations(self):
        design_unit = Entity('tb_entity')
        design_unit.generic_names = ["runner_cfg"]
        test_bench = TestBench(design_unit)

        tests = self.create_tests(test_bench)
        self.assert_has_tests(tests, ["lib.tb_entity.all"])

        test_bench.add_config(name="config", generics=dict())
        tests = self.create_tests(test_bench)
        self.assert_has_tests(tests, ["lib.tb_entity.config"])

    def test_that_pragma_run_in_same_simulation_works(self):
        design_unit = Entity('tb_entity', contents='''\
-- vunit_pragma run_all_in_same_sim
if run("Test_1")
if run("Test_2")
--if run("Test_3")
''')
        design_unit.generic_names = ["runner_cfg"]
        test_bench = TestBench(design_unit)
        tests = self.create_tests(test_bench)
        self.assert_has_tests(tests,
                              [("lib.tb_entity", ("lib.tb_entity.Test_1", "lib.tb_entity.Test_2"))])

    def test_add_config(self):
        design_unit = Entity('tb_entity')
        design_unit.generic_names = ["runner_cfg", "value", "global_value"]
        test_bench = TestBench(design_unit)

        test_bench.set_generic("global_value", "global value")

        test_bench.add_config(name="value=1",
                              generics=dict(value=1,
                                            global_value="local value"))
        test_bench.add_config(name="value=2",
                              generics=dict(value=2))

        tests = self.create_tests(test_bench)
        self.assert_has_tests(tests, ["lib.tb_entity.value=1",
                                      "lib.tb_entity.value=2"])

        self.assertEqual(get_config_of(tests, "lib.tb_entity.value=1").generics,
                         {"value": 1, "global_value": "local value"})
        self.assertEqual(get_config_of(tests, "lib.tb_entity.value=2").generics,
                         {"value": 2, "global_value": "global value"})

    def test_test_case_add_config(self):
        design_unit = Entity('tb_entity', contents='''
if run("test 1")
if run("test 2")
        ''')
        design_unit.generic_names = ["runner_cfg", "value", "global_value"]
        test_bench = TestBench(design_unit)
        test_bench.set_generic("global_value", "global value")
        test_bench.set_sim_option("disable_ieee_warnings", True)

        test_case = test_bench.get_test_case('test 2')
        test_case.add_config(name="c1",
                             generics=dict(value=1,
                                           global_value="local value"))
        test_case.add_config(name="c2",
                             generics=dict(value=2),
                             sim_options=dict(disable_ieee_warnings=False))

        tests = self.create_tests(test_bench)
        self.assert_has_tests(tests, ["lib.tb_entity.test 1",
                                      "lib.tb_entity.c1.test 2",
                                      "lib.tb_entity.c2.test 2"])

        config_test1 = get_config_of(tests, "lib.tb_entity.test 1")
        config_c1_test2 = get_config_of(tests, "lib.tb_entity.c1.test 2")
        config_c2_test2 = get_config_of(tests, "lib.tb_entity.c2.test 2")
        self.assertEqual(config_test1.generics,
                         {"global_value": "global value"})
        self.assertEqual(config_c1_test2.generics,
                         {"value": 1, "global_value": "local value"})
        self.assertEqual(config_c2_test2.generics,
                         {"value": 2, "global_value": "global value"})

    def test_runtime_error_on_configuration_of_individual_test_with_same_sim(self):
        design_unit = Entity('tb_entity', contents='''\
if run("Test 1")
if run("Test 2")
-- vunit_pragma run_all_in_same_sim
''')
        design_unit.generic_names = ["runner_cfg", "name"]
        test_bench = TestBench(design_unit)
        test_case = test_bench.get_test_case("Test 1")
        self.assertRaises(RuntimeError, test_case.set_generic, "name", "value")
        self.assertRaises(RuntimeError, test_case.set_sim_option, "name", "value")
        self.assertRaises(RuntimeError, test_case.add_config, "name")

    def test_run_all_in_same_sim_can_be_configured(self):
        design_unit = Entity('tb_entity', contents='''\
if run("Test 1")
if run("Test 2")
-- vunit_pragma run_all_in_same_sim
''')
        design_unit.generic_names = ["runner_cfg", "name"]
        test_bench = TestBench(design_unit)
        test_bench.set_generic("name", "value")
        test_bench.add_config("cfg")
        tests = self.create_tests(test_bench)
        self.assert_has_tests(tests,
                              [("lib.tb_entity.cfg",
                                ("lib.tb_entity.cfg.Test 1", "lib.tb_entity.cfg.Test 2"))])
        self.assertEqual(get_config_of(tests, "lib.tb_entity.cfg").generics,
                         {"name": "value"})

    def test_fail_on_unknown_sim_option(self):
        design_unit = Entity('tb_entity')
        design_unit.generic_names = ["runner_cfg"]
        test_bench = TestBench(design_unit)
        self.assertRaises(ValueError, test_bench.set_sim_option, "unknown", "value")

    def assert_has_tests(self, test_list, tests):
        """
        Asser that the test_list contains tests.
        A test can be either a string to represent a single test or a
        tuple to represent multiple tests within a test suite.
        """
        self.assertEqual(len(test_list), len(tests))
        for test1, test2 in zip(test_list, tests):
            if isinstance(test2, tuple):
                name, test_cases = test2
                self.assertEqual(test1.name, name)
                self.assertEqual(test1.test_cases, list(test_cases))
            else:
                self.assertEqual(test1.name, test2)

    def create_tests(self, test_bench):
        """
        Helper method to reduce boiler plate
        """
        return test_bench.create_tests(self.simulator_if, elaborate_only=False)


def out(*args):
    return join(dirname(__file__), "test_bench_out", *args)


def get_config_of(tests, test_name):
    """
    Find generic values of test
    """
    for test in tests:
        if test.name == test_name:
            try:
                return test._test_case._config  # pylint: disable=protected-access
            except AttributeError:
                return test._config  # pylint: disable=protected-access
    raise KeyError(test_name)


class Module(object):
    """
    Mock Module
    """
    def __init__(self, name, contents=''):
        self.name = name
        self.library_name = "lib"
        self.is_entity = False
        self.is_module = True
        self.generic_names = []
        self.file_name = out('file.v')
        write_file(out('file.v'), contents)


class Entity(object):  # pylint: disable=too-many-instance-attributes
    """
    Mock Entity
    """
    def __init__(self, name, contents='', no_arch=False):
        self.name = name
        self.library_name = "lib"
        self.is_entity = True
        self.is_module = False
        self.architecture_names = {}

        if not no_arch:
            self.architecture_names = {"arch": out("file.vhd")}

        self.generic_names = []
        self.file_name = out('file.vhd')
        write_file(out('file.vhd'), contents)
        self._add_architecture_callback = None

    def set_add_architecture_callback(self, callback):
        """
        Set callback to be called when an architecture is added
        """
        assert self._add_architecture_callback is None
        self._add_architecture_callback = callback

    def add_architecture(self, name, contents=""):
        """
        Add architecture to this entity
        """
        file_name = name + ".vhd"
        self.architecture_names[name] = out(file_name)
        write_file(out(file_name), contents)

        if self._add_architecture_callback is not None:
            self._add_architecture_callback()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2015-2017, Lars Asplund lars.anders.asplund@gmail.com

"""
Generic simulator interface
"""

from __future__ import print_function
import sys
import os
import subprocess
from vunit.ostools import Process, simplify_path
from vunit.exceptions import CompileError
from vunit.color_printer import NO_COLOR_PRINTER


class SimulatorInterface(object):
    """
    Generic simulator interface
    """

    name = None
    supports_gui_flag = False
    package_users_depend_on_bodies = False
    compile_options = []
    sim_options = []

    def __init__(self):
        self.output_path = None

    @staticmethod
    def add_arguments(parser):
        """
        Add command line arguments
        """
        pass

    @staticmethod
    def supports_vhdl_2008_contexts():
        """
        Returns True when this simulator supports VHDL 2008 contexts
        """
        return True

    @staticmethod
    def find_executable(executable):
        """
        Return a list of all executables found in PATH
        """
        path = os.environ['PATH']
        paths = path.split(os.pathsep)
        _, ext = os.path.splitext(executable)

        if (sys.platform == 'win32' or os.name == 'os2') and (ext != '.exe'):
            executable = executable + '.exe'

        result = []
        if isfile(executable):
            result.append(executable)

        for prefix in paths:
            file_name = os.path.join(prefix, executable)
            if isfile(file_name):
                # the file exists, we have a shot at spawn working
                result.append(file_name)
        return result

    @classmethod
    def find_prefix(cls):
        """
        Find prefix by looking at VUNIT_<SIMULATOR_NAME>_PATH environment variable
        """
        prefix = os.environ.get("VUNIT_" + cls.name.upper() + "_PATH", None)
        if prefix is not None:
            return prefix
        return cls.find_prefix_from_path()

    @classmethod
    def find_prefix_from_path(cls):
        """
        Find simulator toolchain prefix from PATH environment variable
        """
        return None

    @classmethod
    def is_available(cls):
        """
        Returns True if simulator is available
        """
        return cls.find_prefix() is not None

    @classmethod
    def find_toolchain(cls, executables, constraints=None):
        """
        Find the first path prefix containing all executables
        """
        constraints = [] if constraints is None else constraints

        if not executables:
            return None

        all_paths = [[os.path.abspath(os.path.dirname(executables))
                      for executables in cls.find_executable(name)]
                     for name in executables]

        for path0 in all_paths[0]:
            if all([path0 in paths for paths in all_paths] +
                   [constraint(path0) for constraint in constraints]):
                return path0
        return None

    @classmethod
    def get_osvvm_coverage_api(cls):
        """
        Returns simulator name when OSVVM coverage API is supported, None otherwise.
        """
        return None

    @classmethod
    def supports_vhdl_package_generics(cls):
        """
        Returns True when this simulator supports VHDL package generics
        """
        return False

    def post_process(self, output_path):
        """
        Hook for simulator interface to perform post processing such as creating coverage reports
        """
        pass

    def add_simulator_specific(self, project):
        """
        Hook for the simulator interface to add simulator specific things to the project
        """
        pass

    def compile_project(self, project, printer=NO_COLOR_PRINTER, continue_on_error=False):
        """
        Compile the project
        """
        self.add_simulator_specific(project)
        self.setup_library_mapping(project)
        self.compile_source_files(project, printer, continue_on_error)

    def simulate(self, output_path, test_suite_name, config, elaborate_only):
        """
        Simulate
        """
        pass

    def setup_library_mapping(self, project):
        """
        Implemented by specific simulators
        """
        pass

    def __compile_source_file(self, source_file, printer):
        """
        Compiles a single source file and prints status information
        """
        try:
            command = self.compile_source_file_command(source_file)
        except CompileError:
            command = None
            printer.write("failed", fg="ri")
            printer.write("\n")
            printer.write("File type not supported by %s simulator\n" % (self.name))

            return False

        try:
            output = check_output(command,
                                  env=self.get_env())
            printer.write("passed", fg="gi")
            printer.write("\n")
            printer.write(output)

        except subprocess.CalledProcessError as err:
            printer.write("failed", fg="ri")
            printer.write("\n")
            printer.write("=== Command used: ===\n%s\n"
                          % (subprocess.list2cmdline(command)))
            printer.write("\n")
            printer.write("=== Command output: ===\n%s\n" % err.output)

            return False

        return True

    def compile_source_files(self, project, printer=NO_COLOR_PRINTER, continue_on_error=False):
        """
        Use compile_source_file_command to compile all source_files
        """
        dependency_graph = project.create_dependency_graph()
        failures = []
        source_files = project.get_files_in_compile_order(dependency_graph=dependency_graph)
        source_files_to_skip = set()

        max_library_name = 0
        max_source_file_name = 0
        if source_files:
            max_library_name = max(len(source_file.library.name) for source_file in source_files)
            max_source_file_name = max(len(simplify_path(source_file.name)) for source_file in source_files)

        for source_file in source_files:
            printer.write(
                'Compiling into %s %s ' % (
                    (source_file.library.name + ":").ljust(max_library_name + 1),
                    simplify_path(source_file.name).ljust(max_source_file_name)))
            sys.stdout.flush()

            if source_file in source_files_to_skip:
                printer.write("skipped", fg="rgi")
                printer.write("\n")
                continue

            if self.__compile_source_file(source_file, printer):
                project.update(source_file)
            else:
                source_files_to_skip.update(dependency_graph.get_dependent([source_file]))
                failures.append(source_file)

                if not continue_on_error:
                    break

        if failures:
            printer.write("Compile failed\n", fg='ri')
            raise CompileError

        if source_files:
            printer.write("Compile passed\n", fg='gi')
        else:
            printer.write("Re-compile not needed\n")

    def compile_source_file_command(self, source_file):  # pylint: disable=unused-argument
        raise NotImplementedError

    def set_output_path(self, output_path):
        self.output_path = output_path

    def get_output_path(self):
        return self.output_path

    @staticmethod
    def get_env():
        """
        Allows inheriting classes to overload this to modify environment variables
        """
        return None  # Default environment


def isfile(file_name):
    """
    Case insensitive os.path.isfile
    """
    if not os.path.isfile(file_name):
        return False

    return os.path.basename(file_name) in os.listdir(os.path.dirname(file_name))


def run_command(command, cwd=None, env=None):
    """
    Run a command
    """
    try:
        proc = Process(command, cwd=cwd, env=env)
        proc.consume_output()
        return True
    except Process.NonZeroExitCode:
        pass
    return False


def check_output(command, env=None):
    """
    Wrapper arround subprocess.check_output
    """
    try:
        output = subprocess.check_output(command,
                                         env=env,
                                         stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as err:
        err.output = err.output.decode("utf-8")
        raise err
    return output.decode("utf-8")


class Option(object):
    """
    A compile or sim option
    """

    def __init__(self, name):
        self._name = name

    @property
    def name(self):
        return self._name

    def validate(self, value):
        pass


class BooleanOption(Option):
    """
    Must be a boolean
    """

    def validate(self, value):
        if value not in (True, False):
            raise ValueError("Option %r must be a boolean. Got %r"
                             % (self.name, value))


class StringOption(Option):
    """
    Must be a string
    """

    def validate(self, value):
        if not is_string_not_iterable(value):
            raise ValueError("Option %r must be a string. Got %r"
                             % (self.name, value))


class ListOfStringOption(Option):
    """
    Must be a list of strings
    """
    def validate(self, value):
        def fail():
            raise ValueError("Option %r must be a list of strings. Got %r"
                             % (self.name, value))

        if is_string_not_iterable(value):
            fail()

        try:
            for elem in value:
                if not is_string_not_iterable(elem):
                    fail()
        except TypeError:
            fail()


class VHDLAssertLevelOption(Option):
    """
    VHDL assert level
    """

    _legal_values = ("warning", "error", "failure")

    def __init__(self):
        Option.__init__(self, "vhdl_assert_stop_level")

    def validate(self, value):
        if value not in self._legal_values:
            raise ValueError("Option %r must be one of %s. Got %r"
                             % (self.name, self._legal_values, value))


def is_string_not_iterable(value):
    """
    Returns True if value is a string and not another iterable
    """
    if sys.version_info.major == 3:
        return isinstance(value, str)

    return isinstance(value, (str, unicode))  # pylint: disable=undefined-variable



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2016-2017, Lars Asplund lars.anders.asplund@gmail.com

"""
Test the ActiveHDL interface
"""


import unittest
from os.path import join, dirname, exists
import os
from shutil import rmtree
from vunit.activehdl_interface import ActiveHDLInterface
from vunit.test.mock_2or3 import mock
from vunit.project import Project
from vunit.ostools import renew_path, write_file


class TestActiveHDLInterface(unittest.TestCase):
    """
    Test the ActiveHDL interface
    """

    @mock.patch("vunit.simulator_interface.check_output", autospec=True, return_value="")
    @mock.patch("vunit.activehdl_interface.Process", autospec=True)
    def test_compile_project_vhdl(self, process, check_output):
        library_cfg = join(self.output_path, "library.cfg")
        simif = ActiveHDLInterface(prefix="prefix",
                                   library_cfg=library_cfg)
        project = Project()
        project.add_library("lib", "lib_path")
        write_file("file.vhd", "")
        project.add_source_file("file.vhd", "lib", file_type="vhdl")
        simif.compile_project(project)
        process.assert_any_call([join("prefix", "vlib"), "lib", "lib_path"],
                                cwd=self.output_path,
                                env=simif.get_env())
        process.assert_called_with([join("prefix", "vmap"), "lib", "lib_path"],
                                   cwd=self.output_path,
                                   env=simif.get_env())
        check_output.assert_called_once_with(
            [join('prefix', 'vcom'),
             '-quiet',
             '-j',
             self.output_path,
             '-2008',
             '-work',
             'lib',
             'file.vhd'],
            env=simif.get_env())

    @mock.patch("vunit.simulator_interface.check_output", autospec=True, return_value="")
    @mock.patch("vunit.activehdl_interface.Process", autospec=True)
    def test_compile_project_vhdl_extra_flags(self, process, check_output):
        library_cfg = join(self.output_path, "library.cfg")
        simif = ActiveHDLInterface(prefix="prefix",
                                   library_cfg=library_cfg)
        project = Project()
        project.add_library("lib", "lib_path")
        write_file("file.vhd", "")
        source_file = project.add_source_file("file.vhd", "lib", file_type="vhdl")
        source_file.set_compile_option("activehdl.vcom_flags", ["custom", "flags"])
        simif.compile_project(project)
        process.assert_any_call([join("prefix", "vlib"), "lib", "lib_path"],
                                cwd=self.output_path,
                                env=simif.get_env())
        process.assert_called_with([join("prefix", "vmap"), "lib", "lib_path"],
                                   cwd=self.output_path,
                                   env=simif.get_env())
        check_output.assert_called_once_with([join('prefix', 'vcom'),
                                              '-quiet',
                                              '-j',
                                              self.output_path,
                                              'custom',
                                              'flags',
                                              '-2008',
                                              '-work',
                                              'lib',
                                              'file.vhd'],
                                             env=simif.get_env())

    @mock.patch("vunit.simulator_interface.check_output", autospec=True, return_value="")
    @mock.patch("vunit.activehdl_interface.Process", autospec=True)
    def test_compile_project_verilog(self, process, check_output):
        library_cfg = join(self.output_path, "library.cfg")
        simif = ActiveHDLInterface(prefix="prefix",
                                   library_cfg=library_cfg)
        project = Project()
        project.add_library("lib", "lib_path")
        write_file("file.v", "")
        project.add_source_file("file.v", "lib", file_type="verilog")
        simif.compile_project(project)
        process.assert_any_call([join("prefix", "vlib"), "lib", "lib_path"],
                                cwd=self.output_path,
                                env=simif.get_env())
        process.assert_called_with([join("prefix", "vmap"), "lib", "lib_path"],
                                   cwd=self.output_path,
                                   env=simif.get_env())
        check_output.assert_called_once_with([join('prefix', 'vlog'),
                                              '-quiet',
                                              '-lc',
                                              library_cfg,
                                              '-work',
                                              'lib',
                                              'file.v',
                                              '-l', 'lib'],
                                             env=simif.get_env())

    @mock.patch("vunit.simulator_interface.check_output", autospec=True, return_value="")
    @mock.patch("vunit.activehdl_interface.Process", autospec=True)
    def test_compile_project_system_verilog(self, process, check_output):
        library_cfg = join(self.output_path, "library.cfg")
        simif = ActiveHDLInterface(prefix="prefix",
                                   library_cfg=library_cfg)
        project = Project()
        project.add_library("lib", "lib_path")
        write_file("file.sv", "")
        project.add_source_file("file.sv", "lib", file_type="systemverilog")
        simif.compile_project(project)
        process.assert_any_call([join("prefix", "vlib"), "lib", "lib_path"],
                                cwd=self.output_path,
                                env=simif.get_env())
        process.assert_called_with([join("prefix", "vmap"), "lib", "lib_path"],
                                   cwd=self.output_path,
                                   env=simif.get_env())
        check_output.assert_called_once_with([join('prefix', 'vlog'),
                                              '-quiet',
                                              '-lc',
                                              library_cfg,
                                              '-work',
                                              'lib',
                                              'file.sv',
                                              '-l', 'lib'],
                                             env=simif.get_env())

    @mock.patch("vunit.simulator_interface.check_output", autospec=True, return_value="")
    @mock.patch("vunit.activehdl_interface.Process", autospec=True)
    def test_compile_project_verilog_extra_flags(self, process, check_output):
        library_cfg = join(self.output_path, "library.cfg")
        simif = ActiveHDLInterface(prefix="prefix",
                                   library_cfg=library_cfg)
        project = Project()
        project.add_library("lib", "lib_path")
        write_file("file.v", "")
        source_file = project.add_source_file("file.v", "lib", file_type="verilog")
        source_file.set_compile_option("activehdl.vlog_flags", ["custom", "flags"])
        simif.compile_project(project)
        process.assert_any_call([join("prefix", "vlib"), "lib", "lib_path"],
                                cwd=self.output_path,
                                env=simif.get_env())
        process.assert_called_with([join("prefix", "vmap"), "lib", "lib_path"],
                                   cwd=self.output_path,
                                   env=simif.get_env())
        check_output.assert_called_once_with([join('prefix', 'vlog'),
                                              '-quiet',
                                              '-lc',
                                              library_cfg,
                                              'custom',
                                              'flags',
                                              '-work',
                                              'lib',
                                              'file.v',
                                              '-l', 'lib'],
                                             env=simif.get_env())

    @mock.patch("vunit.simulator_interface.check_output", autospec=True, return_value="")
    @mock.patch("vunit.activehdl_interface.Process", autospec=True)
    def test_compile_project_verilog_include(self, process, check_output):
        library_cfg = join(self.output_path, "library.cfg")
        simif = ActiveHDLInterface(prefix="prefix",
                                   library_cfg=library_cfg)
        project = Project()
        project.add_library("lib", "lib_path")
        write_file("file.v", "")
        project.add_source_file("file.v", "lib", file_type="verilog", include_dirs=["include"])
        simif.compile_project(project)
        process.assert_any_call([join("prefix", "vlib"), "lib", "lib_path"],
                                cwd=self.output_path,
                                env=simif.get_env())
        process.assert_called_with([join("prefix", "vmap"), "lib", "lib_path"],
                                   cwd=self.output_path,
                                   env=simif.get_env())
        check_output.assert_called_once_with([join('prefix', 'vlog'),
                                              '-quiet',
                                              '-lc',
                                              library_cfg,
                                              '-work',
                                              'lib',
                                              'file.v',
                                              '-l', 'lib',
                                              '+incdir+include'],
                                             env=simif.get_env())

    @mock.patch("vunit.simulator_interface.check_output", autospec=True, return_value="")
    @mock.patch("vunit.activehdl_interface.Process", autospec=True)
    def test_compile_project_verilog_define(self, process, check_output):
        library_cfg = join(self.output_path, "library.cfg")
        simif = ActiveHDLInterface(prefix="prefix",
                                   library_cfg=library_cfg)
        project = Project()
        project.add_library("lib", "lib_path")
        write_file("file.v", "")
        project.add_source_file("file.v", "lib", file_type="verilog", defines={"defname": "defval"})
        simif.compile_project(project)
        process.assert_any_call([join("prefix", "vlib"), "lib", "lib_path"],
                                cwd=self.output_path,
                                env=simif.get_env())
        process.assert_called_with([join("prefix", "vmap"), "lib", "lib_path"],
                                   cwd=self.output_path,
                                   env=simif.get_env())
        check_output.assert_called_once_with([join('prefix', 'vlog'),
                                              '-quiet',
                                              '-lc',
                                              library_cfg,
                                              '-work',
                                              'lib',
                                              'file.v',
                                              '-l', 'lib',
                                              '+define+defname=defval'],
                                             env=simif.get_env())

    def setUp(self):
        self.output_path = join(dirname(__file__), "test_activehdl_out")
        renew_path(self.output_path)
        self.project = Project()
        self.cwd = os.getcwd()
        os.chdir(self.output_path)

    def tearDown(self):
        os.chdir(self.cwd)
        if exists(self.output_path):
            rmtree(self.output_path)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2016-2017, Lars Asplund lars.anders.asplund@gmail.com

"""
Interface for the Cadence Incisive simulator
"""

from __future__ import print_function
import os
from os.path import join, dirname, abspath, relpath
import subprocess
import sys
import logging
from vunit.ostools import write_file, file_exists
from vunit.simulator_interface import (SimulatorInterface,
                                       run_command,
                                       ListOfStringOption)
from vunit.exceptions import CompileError
from vunit.cds_file import CDSFile
LOGGER = logging.getLogger(__name__)


class IncisiveInterface(SimulatorInterface):  # pylint: disable=too-many-instance-attributes
    """
    Interface for the Cadence Incisive simulator
    """

    name = "incisive"
    supports_gui_flag = True
    package_users_depend_on_bodies = False

    compile_options = [
        ListOfStringOption("incisive.irun_vhdl_flags"),
        ListOfStringOption("incisive.irun_verilog_flags"),
    ]

    sim_options = [
        ListOfStringOption("incisive.irun_sim_flags")
    ]

    @staticmethod
    def add_arguments(parser):
        """
        Add command line arguments
        """
        group = parser.add_argument_group("Incisive irun",
                                          description="Incisive irun-specific flags")
        group.add_argument("--cdslib",
                           default=None,
                           help="The cds.lib file to use. If not given, VUnit maintains its own cds.lib file.")
        group.add_argument("--hdlvar",
                           default=None,
                           help="The hdl.var file to use. If not given, VUnit does not use a hdl.var file.")

    @classmethod
    def from_args(cls, output_path, args):
        """
        Create new instance from command line arguments object
        """
        return cls(prefix=cls.find_prefix(),
                   output_path=output_path,
                   log_level=args.log_level,
                   gui=args.gui,
                   cdslib=args.cdslib,
                   hdlvar=args.hdlvar)

    @classmethod
    def find_prefix_from_path(cls):
        """
        Find incisive simulator from PATH environment variable
        """
        return cls.find_toolchain(['irun'])

    @staticmethod
    def supports_vhdl_2008_contexts():
        """
        Returns True when this simulator supports VHDL 2008 contexts
        """
        return False

    def __init__(self,  # pylint: disable=too-many-arguments
                 prefix, output_path, gui=False, log_level=None, cdslib=None, hdlvar=None):
        SimulatorInterface.__init__(self)
        self._prefix = prefix
        self._libraries = []
        self._output_path = output_path
        self._gui = gui
        self._log_level = log_level
        if cdslib is None:
            self._cdslib = abspath(join(output_path, 'cds.lib'))
        else:
            self._cdslib = abspath(cdslib)
        self._hdlvar = hdlvar
        self._cds_root_irun = self.find_cds_root_irun()
        self._create_cdslib()

    def find_cds_root_irun(self):
        """
        Finds irun cds root
        """
        return subprocess.check_output([join(self._prefix, 'cds_root'), 'irun']).splitlines()[0]

    def find_cds_root_virtuoso(self):
        """
        Finds virtuoso cds root
        """
        try:
            return subprocess.check_output([join(self._prefix, 'cds_root'), 'virtuoso']).splitlines()[0]
        except subprocess.CalledProcessError:
            return None

    def _create_cdslib(self):
        """
        Create the cds.lib file in the output directory if it does not exist
        """
        cds_root_virtuoso = self.find_cds_root_virtuoso()

        if cds_root_virtuoso is None:
            contents = """\
## cds.lib: Defines the locations of compiled libraries.
softinclude {0}/tools/inca/files/cds.lib
# needed for referencing the library 'basic' for cells 'cds_alias', 'cds_thru' etc. in analog models:
# NOTE: 'virtuoso' executable not found!
# define basic ".../tools/dfII/etc/cdslib/basic"
define work "{1}/libraries/work"
""".format(self._cds_root_irun, self._output_path)
        else:
            contents = """\
## cds.lib: Defines the locations of compiled libraries.
softinclude {0}/tools/inca/files/cds.lib
# needed for referencing the library 'basic' for cells 'cds_alias', 'cds_thru' etc. in analog models:
define basic "{1}/tools/dfII/etc/cdslib/basic"
define work "{2}/libraries/work"
""".format(self._cds_root_irun, cds_root_virtuoso, self._output_path)

        write_file(self._cdslib, contents)

    def setup_library_mapping(self, project):
        """
        Compile project using vhdl_standard
        """
        mapped_libraries = self._get_mapped_libraries()

        for library in project.get_libraries():
            self._libraries.append(library)
            self.create_library(library.name, library.directory, mapped_libraries)

    def compile_source_file_command(self, source_file):
        """
        Returns the command to compile a single source file
        """
        if source_file.is_vhdl:
            return self.compile_vhdl_file_command(source_file)
        elif source_file.is_any_verilog:
            return self.compile_verilog_file_command(source_file)

        raise CompileError

    @staticmethod
    def _vhdl_std_opt(vhdl_standard):
        """
        Convert standard to format of irun command line flag
        """
        if vhdl_standard == "2002":
            return "-v200x -extv200x"
        elif vhdl_standard == "2008":
            return "-v200x -extv200x"
        elif vhdl_standard == "93":
            return "-v93"
        raise ValueError("Invalid VHDL standard %s" % vhdl_standard)

    def compile_vhdl_file_command(self, source_file):
        """
        Returns command to compile a VHDL file
        """
        cmd = join(self._prefix, 'irun')
        args = []
        args += ['-compile']
        args += ['-nocopyright']
        args += ['-licqueue']
        args += ['-nowarn DLCPTH']  # "cds.lib Invalid path"
        args += ['-nowarn DLCVAR']  # "cds.lib Invalid environment variable ''."
        args += ['%s' % self._vhdl_std_opt(source_file.get_vhdl_standard())]
        args += ['-work work']
        args += ['-cdslib "%s"' % self._cdslib]
        args += self._hdlvar_args()
        args += ['-log "%s"' % join(self._output_path, "irun_compile_vhdl_file_%s.log" % source_file.library.name)]
        if not self._log_level == "debug":
            args += ['-quiet']
        else:
            args += ['-messages']
            args += ['-libverbose']
        args += source_file.compile_options.get('incisive.irun_vhdl_flags', [])
        args += ['-nclibdirname "%s"' % dirname(source_file.library.directory)]
        args += ['-makelib %s' % source_file.library.directory]
        args += ['"%s"' % source_file.name]
        args += ['-endlib']
        argsfile = join(self._output_path, "irun_compile_vhdl_file_%s.args" % source_file.library.name)
        write_file(argsfile, "\n".join(args))
        return [cmd, '-f', argsfile]

    def compile_verilog_file_command(self, source_file):
        """
        Returns commands to compile a Verilog file
        """
        cmd = join(self._prefix, 'irun')
        args = []
        args += ['-compile']
        args += ['-nocopyright']
        args += ['-licqueue']
        # "Ignored unexpected semicolon following SystemVerilog description keyword (endfunction)."
        args += ['-nowarn UEXPSC']
        # "cds.lib Invalid path"
        args += ['-nowarn DLCPTH']
        # "cds.lib Invalid environment variable ''."
        args += ['-nowarn DLCVAR']
        args += ['-work work']
        args += source_file.compile_options.get('incisive.irun_verilog_flags', [])
        args += ['-cdslib "%s"' % self._cdslib]
        args += self._hdlvar_args()
        args += ['-log "%s"' % join(self._output_path, "irun_compile_verilog_file_%s.log" % source_file.library.name)]
        if not self._log_level == "debug":
            args += ['-quiet']
        else:
            args += ['-messages']
            args += ['-libverbose']
        for include_dir in source_file.include_dirs:
            args += ['-incdir "%s"' % include_dir]

        # for "disciplines.vams" etc.
        args += ['-incdir "%s/tools/spectre/etc/ahdl/"' % self._cds_root_irun]

        for key, value in source_file.defines.items():
            args += ['-define %s=%s' % (key, value.replace('"', '\\"'))]
        args += ['-nclibdirname "%s"' % dirname(source_file.library.directory)]
        args += ['-makelib %s' % source_file.library.name]
        args += ['"%s"' % source_file.name]
        args += ['-endlib']
        argsfile = join(self._output_path, "irun_compile_verilog_file_%s.args" % source_file.library.name)
        write_file(argsfile, "\n".join(args))
        return [cmd, '-f', argsfile]

    def create_library(self, library_name, library_path, mapped_libraries=None):
        """
        Create and map a library_name to library_path
        """
        mapped_libraries = mapped_libraries if mapped_libraries is not None else {}

        if not file_exists(dirname(abspath(library_path))):
            os.makedirs(dirname(abspath(library_path)))

        if library_name in mapped_libraries and mapped_libraries[library_name] == library_path:
            return

        cds = CDSFile.parse(self._cdslib)
        cds[library_name] = library_path
        cds.write(self._cdslib)

    def _get_mapped_libraries(self):
        """
        Get mapped libraries from cds.lib file
        """
        cds = CDSFile.parse(self._cdslib)
        return cds

    def simulate(self,  # pylint: disable=too-many-locals
                 output_path, test_suite_name, config, elaborate_only=False):
        """
        Elaborates and Simulates with entity as top level using generics
        """

        launch_gui = self._gui is not False and not elaborate_only

        if elaborate_only:
            steps = ['elaborate']
        else:
            steps = ['elaborate', 'simulate']

        for step in steps:
            cmd = join(self._prefix, 'irun')
            args = []
            if step == 'elaborate':
                args += ['-elaborate']
            args += ['-nocopyright']
            args += ['-licqueue']
            # args += ['-dumpstack']
            # args += ['-gdbsh']
            # args += ['-rebuild']
            # args += ['-gdb']
            # args += ['-gdbelab']
            args += ['-errormax 10']
            args += ['-nowarn WRMNZD']
            args += ['-nowarn DLCPTH']  # "cds.lib Invalid path"
            args += ['-nowarn DLCVAR']  # "cds.lib Invalid environment variable ''."
            args += ['-ncerror EVBBOL']  # promote to error: "bad boolean literal in generic association"
            args += ['-ncerror EVBSTR']  # promote to error: "bad string literal in generic association"
            args += ['-ncerror EVBNAT']  # promote to error: "bad natural literal in generic association"
            args += ['-work work']
            args += ['-nclibdirname "%s"' % (join(self._output_path, "libraries"))]  # @TODO: ugly
            args += config.sim_options.get('incisive.irun_sim_flags', [])
            args += ['-cdslib "%s"' % self._cdslib]
            args += self._hdlvar_args()
            args += ['-log "%s/irun_%s.log"' % (output_path, step)]
            if not self._log_level == "debug":
                args += ['-quiet']
            else:
                args += ['-messages']
                # args += ['-libverbose']
            args += self._generic_args(config.entity_name, config.generics)
            for library in self._libraries:
                args += ['-reflib "%s"' % library.directory]
            if launch_gui:
                args += ['-access +rwc']
                # args += ['-linedebug']
                args += ['-gui']
            else:
                args += ['-access +r']
                args += ['-input "@run"']
                # Try hierarchical path formats for both VHDL and Verilog, but don't throw an error if not found.
                # args += ['-input "@catch {puts #vunit_pkg::__runner__.exit_without_errors}"']
                # args += ['-input "@catch {puts #run_base_pkg.runner.exit_without_errors}"']
                # NOTE: do not exit with 1 or 2 in case of error, that seems to mean something special to Incisive:
                args += ['-input "@catch '
                         '{if {#vunit_pkg::__runner__.exit_without_errors == 1} {exit 0} else {exit 42}}"']
                args += ['-input "@catch '
                         '{if {#run_base_pkg.runner.exit_without_errors == \\"TRUE\\"} {exit 0} else {exit 42}}"']
            if config.architecture_name is None:
                # we have a SystemVerilog toplevel:
                args += ['-top %s' % join('%s.%s:sv' % (config.library_name, config.entity_name))]
            else:
                # we have a VHDL toplevel:
                args += ['-top %s' % join('%s.%s:%s' % (config.library_name, config.entity_name,
                                                        config.architecture_name))]
            argsfile = "%s/irun_%s.args" % (output_path, step)
            write_file(argsfile, "\n".join(args))
            if not run_command([cmd, '-f', relpath(argsfile, output_path)], cwd=output_path, env=self.get_env()):
                return False
        return True

    def _hdlvar_args(self):
        """
        Return hdlvar argument if available
        """
        if self._hdlvar is None:
            return []
        return ['-hdlvar "%s"' % self._hdlvar]

    @staticmethod
    def _generic_args(entity_name, generics):
        """
        Create irun arguments for generics/parameters
        """
        args = []
        for name, value in generics.items():
            if _generic_needs_quoting(value):
                args += ['''-gpg "%s.%s => \\"%s\\""''' % (entity_name, name, value)]
            else:
                args += ['''-gpg "%s.%s => %s"''' % (entity_name, name, value)]
        return args


def _generic_needs_quoting(value):  # pylint: disable=missing-docstring
    if sys.version_info.major == 2:
        return isinstance(value, (str, bool, unicode))  # pylint: disable=undefined-variable

    return isinstance(value, (str, bool))
